import os
import random
from typing import Union

import cv2
import numpy as np
from tqdm import tqdm

import supervisely as sly
from dataset_tools.image.renders.convert import compress_png


class SideAnnotationsGrid:
    def __init__(
        self,
        project: Union[str, int],
        project_meta: sly.ProjectMeta,
        api: sly.Api = None,
        rows: int = 3,
        cols: int = 3,
        force: bool = False,
    ):
        self.force = force
        self.project_meta = project_meta

        self._img_height = 1080
        self._rows = rows
        self._cols = cols
        self._row_width = 0
        self._gap = 15
        self._bg_color = (221, 210, 230)  # rgb(221, 210, 230)

        self._local = False if isinstance(project, int) else True
        self._api = api if api is not None else sly.Api.from_env()

        self.np_images = []
        self._img_array = None
        self._row_height = int((self._img_height - self._gap * (self._rows + 1)) / self._rows)

        self._side_overlay_path = "side_logo_overlay.png"
        self._logo_path = "logo.png"

    @property
    def basename_stem(self) -> None:
        return sly.utils.camel_to_snake(self.__class__.__name__)

    def update(self, data: tuple):
        cnt = self._cols * self._rows
        join_data = [(ds, img, ann) for ds, list1, list2 in data for img, ann in zip(list1, list2)]

        random.shuffle(join_data)
        with tqdm(desc="Downloading images", total=cnt) as p:
            for ds, img_info, ann in join_data[:cnt]:
                img = (
                    sly.image.read(ds.get_img_path(img_info.name))
                    if self._local
                    else self._api.image.download_np(img_info.id)
                )
                mask = self._draw_masks_on_single_image(ann, img_info)

                img = self._resize_image(img)
                mask = self._resize_image(mask)
                join_image = np.hstack([img, mask])
                self.np_images.append(join_image)

                p.update(1)

    def _draw_masks_on_single_image(self, ann: sly.Annotation, image: sly.ImageInfo):
        height, width = image.height, image.width
        mask_img = np.zeros((height, width, 3), dtype=np.uint8)
        mask_img[:, :, 0:3] = self._bg_color  # rgb(221, 210, 230)

        for label in ann.labels:
            label.geometry.draw(mask_img, color=label.obj_class.color, thickness=3)

        return mask_img

    def to_image(self, path: str = None):
        path_part, ext = os.path.splitext(path)
        tmp_path = f"{path_part}-o{ext}"
        if path is None:
            storage_dir = sly.app.get_data_dir()
            path = os.path.join(storage_dir, "horizontal_grid.png")

        self._merge_canvas_with_images()
        self._add_logo(self._img_array)

        sly.image.write(tmp_path, self._img_array)
        compress_png(tmp_path, path)
        sly.fs.silent_remove(tmp_path)
        sly.logger.info(f"Result grid saved to: {path}")

    def _create_image_canvas(self):
        self._img_array = np.ones([self._img_height, self._row_width, 3], dtype=np.uint8) * 255

    def _merge_canvas_with_images(self):
        rows = self._create_rows()
        rows = self._merge_img_in_rows(rows)
        self._create_image_canvas()
        for i, image in enumerate(rows):
            if i >= self._rows:
                continue
            if image.shape[1] > self._img_array.shape[1] - self._gap:
                image = image[:, : self._img_array.shape[1] - self._gap]

            row_start = i * (self._row_height + self._gap) + self._gap
            row_end = row_start + self._row_height
            column_start = self._gap
            column_end = self._img_array.shape[1]

            self._img_array[row_start:row_end, column_start:column_end] = image

    def _create_rows(self):
        num_images = len(self.np_images)
        image_widths = [image.shape[1] for image in self.np_images]

        one_big_row_width = sum(image_widths) + (num_images - 1) * self._gap
        self._row_width = one_big_row_width // self._rows

        rows = []
        row_images = []
        current_width = 0

        for idx, (image, width) in enumerate(zip(self.np_images, image_widths)):
            if current_width + width > self._row_width:
                rows.append(row_images)

                row_images = []
                current_width = 0

            row_images.append(image)
            if idx == num_images - 1:
                rows.append(row_images)
            current_width += width + self._gap

        if len(rows) == self._rows:
            return rows
        return rows

    def _merge_img_in_rows(self, rows):
        combined_rows = []
        separator = np.ones((self._row_height, 15, 3), dtype=np.uint8) * 255
        for row in rows:
            combined_images = []

            for image in row:
                combined_images.append(image)
                combined_images.append(separator)
            combined_images.pop()
            combined_image = np.hstack(combined_images)
            combined_rows.append(combined_image)

        self._row_width = min([img.shape[1] for img in combined_rows])
        return combined_rows

    def _resize_image(self, image, height=None):
        img_h, img_w = image.shape[:2]
        if height is None:
            height = self._row_height
        img_aspect_ratio = height / img_h
        img_h = int(height)
        img_w = int(img_aspect_ratio * img_w)

        image = sly.image.resize(image, (img_h, img_w))

        return image

    def _add_overlay_with_logo(self):
        image2 = cv2.imread(self._side_overlay_path, cv2.IMREAD_UNCHANGED)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGRA2RGBA)

        _, width1 = self._img_array.shape[:2]
        height2, width2 = image2.shape[:2]

        alpha_channel = image2[:, :, 3] / 255.0

        x = width1 - width2

        region = self._img_array[:height2, x : x + width2]
        self._img_array[:height2, x : x + width2, :3] = (
            1 - alpha_channel[:, :, np.newaxis]
        ) * region + alpha_channel[:, :, np.newaxis] * image2[:, :, :3]


    def _add_logo(self, image):
        height = max(self._row_height // 5, 80)
        image2 = cv2.imread(self._logo_path, cv2.IMREAD_UNCHANGED)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGRA2RGBA)
        image2 = self._resize_image(image2, height)

        height_r = height
        while image2.shape[1] > image.shape[1] * 0.25:
            height_r = int(0.95 * height_r)
            image2 = self._resize_image(image2, height_r)

        h1, w1 = image.shape[:2]
        h2, w2 = image2.shape[:2]
        x = w1 - w2 - self._gap * 2
        y = h1 - h2 - self._gap * 2
        alpha = image2[:, :, 3] / 255.0

        reg = image[y : y + h2, x : x + w2]
        image[y : y + h2, x : x + w2, :3] = (1 - alpha[:, :, np.newaxis]) * reg[:, :, :3] + alpha[
            :, :, np.newaxis
        ] * image2[:, :, :3]

        return image



import os
import random
from typing import Union

import cv2
import numpy as np
from tqdm import tqdm

import supervisely as sly
from dataset_tools.image.renders.convert import from_mp4_to_webm, compress_mp4, compress_png


class HorizontalGrid:
    def __init__(
        self,
        project: Union[str, int],
        project_meta: sly.ProjectMeta,
        api: sly.Api = None,
        rows: int = 3,
        cols: int = 6,
        force: bool = False,
    ):
        self.force = force
        self.project_meta = project_meta

        self._img_height = 1920
        self._rows = rows
        self._cols = cols
        self._gap = 15
        self._row_width = 0
        self._logo_path = "logo.png"

        self.np_images = []  # for grid
        self.np_anns = []  # for gif
        self.np_frames = []  # for gif
        self._img_array = None
        self._row_height = int((self._img_height - self._gap * (self._rows + 1)) / self._rows)

        self._local = False if isinstance(project, int) else True
        self._api = api if api is not None else sly.Api.from_env()

    @property
    def basename_stem(self) -> None:
        return sly.utils.camel_to_snake(self.__class__.__name__)

    def update(self, data: tuple):
        cnt = self._cols * self._rows
        join_data = [(ds, img, ann) for ds, list1, list2 in data for img, ann in zip(list1, list2)]

        random.shuffle(join_data)
        with tqdm(desc="Downloading images", total=cnt) as p:
            for ds, img_info, ann in join_data[:cnt]:
                img = (
                    sly.image.read(ds.get_img_path(img_info.name))
                    if self._local
                    else self._api.image.download_np(img_info.id)
                )
                p.update(1)
                ann: sly.Annotation
                tmp = np.dstack((img, np.ones((*img.shape[:2], 1), dtype=np.uint8) * 255))
                ann_mask = np.ones((*img.shape[:2], 4), dtype=np.uint8) * 255
                ann.draw_pretty(ann_mask[:, :, :3], thickness=0, opacity=0.7)
                self.np_frames.append(self._resize_image(tmp, self._row_height))  # for gif
                self.np_anns.append(self._resize_image(ann_mask, self._row_height))  # for gif

                ann.draw_pretty(img, thickness=0, opacity=0.7)  # for grid
                img = self._resize_image(img, self._row_height)
                self.np_images.append(img)

    def to_image(self, path: str = None):
        path_part, ext = os.path.splitext(path)
        tmp_path = f"{path_part}-o{ext}"
        if path is None:
            storage_dir = sly.app.get_data_dir()
            path = os.path.join(storage_dir, "horizontal_grid.png")

        self._img_array = self._merge_canvas_with_images(self.np_images)
        self._add_logo(self._img_array)

        sly.image.write(tmp_path, self._img_array)
        compress_png(tmp_path, path)
        sly.fs.silent_remove(tmp_path)
        sly.logger.info(f"Result grid saved to: {path}")

    def animate(self, path: str = None):
        bg = self._merge_canvas_with_images(self.np_frames, 4)
        ann = self._merge_canvas_with_images(self.np_anns, 4)
        ann[:, :, 3] = np.where(np.all(ann == 255, axis=-1), 0, 255).astype(np.uint8)

        duration = 1.1
        fps = 15
        num_frames = int(duration * fps)
        frames = []
        for i in list(range(1, num_frames + 1)) + list(range(num_frames, 0, -1)):
            alpha = i / num_frames
            frame = self._overlay_images(bg, ann, alpha)
            self._add_logo(frame)
            frame = self._resize_image(frame, frame.shape[0] // 2)
            frames.append(frame)
            if i == num_frames:
                frames.extend([frame] * (fps // 2))

        tmp_video_path = f"{os.path.splitext(path)[0]}-o.mp4"
        video_path = f"{os.path.splitext(path)[0]}.mp4"
        self._save_video(tmp_video_path, frames)
        from_mp4_to_webm(tmp_video_path, path)
        compress_mp4(tmp_video_path, video_path)
        sly.fs.silent_remove(tmp_video_path)

        sly.logger.info(f"Animation saved to: {path}, {video_path}")

    def _merge_canvas_with_images(self, images, channels: int = 3):
        rows = self._create_rows(images)
        rows = self._merge_img_in_rows(rows, channels)

        canvas = np.ones([self._img_height, self._row_width, channels], dtype=np.uint8) * 255
        for i, image in enumerate(rows):
            if image.shape[1] > canvas.shape[1] - self._gap * 2:
                image = image[:, : canvas.shape[1] - self._gap * 2]

            row_start = i * (self._row_height + self._gap) + self._gap
            row_end = row_start + self._row_height
            column_start = self._gap
            column_end = image.shape[1] + self._gap

            if row_end > canvas.shape[0]:
                continue
            canvas[row_start:row_end, column_start:column_end] = image
        return canvas

    def _create_rows(self, images):
        num_images = len(images)
        image_widths = [image.shape[1] for image in images]

        one_big_row_width = sum(image_widths[: -self._rows]) + (num_images - 1) * self._gap
        self._row_width = one_big_row_width // self._rows

        rows = []
        row_images = []
        current_width = 0

        for idx, (image, width) in enumerate(zip(images, image_widths)):
            if current_width + width > self._row_width or idx == len(images) - 1:
                rows.append(row_images)

                row_images = []
                current_width = 0
            row_images.append(image)
            current_width += width + self._gap

        return rows

    def _merge_img_in_rows(self, rows, channels: int = 3):
        max_width = 0
        img_widths = []
        for row in rows:
            sum_widths = sum([img.shape[1] for img in row])
            img_widths.append(sum_widths)
            min_gap = self._gap * (len(row) + 1)
            max_width = max(max_width, sum_widths + min_gap)
        self._row_width = max_width

        combined_rows = []

        for row, width in zip(rows, img_widths):
            gap = (self._row_width - width) // max((len(row) - 1), 1)
            separator = np.ones((self._row_height, gap, channels), dtype=np.uint8) * 255
            combined_images = []

            for image in row:
                combined_images.append(image)
                combined_images.append(separator)
            combined_images.pop()
            combined_image = np.hstack(combined_images)
            combined_rows.append(combined_image)

        return combined_rows

    def _resize_image(self, image, height):
        img_h, img_w = image.shape[:2]
        img_aspect_ratio = height / img_h
        img_h = height
        img_w = int(img_aspect_ratio * img_w)

        image = sly.image.resize(image, (img_h, img_w))

        return image

    def _add_logo(self, image):
        height = max(self._row_height // 5, 80)
        image2 = cv2.imread(self._logo_path, cv2.IMREAD_UNCHANGED)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGRA2RGBA)
        image2 = self._resize_image(image2, height)

        h1, w1 = image.shape[:2]
        h2, w2 = image2.shape[:2]
        x = w1 - w2 - self._gap * 2
        y = h1 - h2 - self._gap * 2
        alpha = image2[:, :, 3] / 255.0

        reg = image[y : y + h2, x : x + w2]
        image[y : y + h2, x : x + w2, :3] = (1 - alpha[:, :, np.newaxis]) * reg[:, :, :3] + alpha[
            :, :, np.newaxis
        ] * image2[:, :, :3]

    def _overlay_images(self, bg, overlay, opacity):
        alpha = overlay[..., 3] * opacity / 255.0
        one_minus_alpha = 1 - alpha

        result_image = np.copy(bg)
        for c in range(3):
            result_image[..., c] = bg[..., c] * one_minus_alpha + overlay[..., c] * alpha
        return result_image

    def _save_video(self, videopath: str, frames):
        fourcc = cv2.VideoWriter_fourcc(*"VP90")
        height, width = frames[0].shape[:2]
        video_writer = cv2.VideoWriter(videopath, fourcc, 15, (width, height))

        with tqdm(desc="Saving video...", total=len(frames)) as vid_pbar:
            for frame in frames:
                if frame.shape[:2] != (height, width):
                    raise Exception("Not all frame sizes are not equal to each other.")
                frame = cv2.cvtColor(frame, cv2.COLOR_RGBA2BGR)
                video_writer.write(frame)
                vid_pbar.update(1)
        video_writer.release()



from dataset_tools.download import prepare_download_link, update_sly_url_dict

from dataset_tools.image.stats.class_balance import ClassBalance
from dataset_tools.image.stats.class_cooccurrence import ClassCooccurrence

from dataset_tools.image.stats.classes_per_image import ClassesPerImage
from dataset_tools.image.stats.heatmaps_for_classes import ClassesHeatmaps
from dataset_tools.image.stats.preview_for_classes import ClassesPreview

from dataset_tools.image.stats.objects_distribution import ObjectsDistribution
from dataset_tools.image.stats.object_and_class_sizes import ObjectSizes, ClassSizes

from dataset_tools.image.stats.wrapper import count_stats  # , initialize

from dataset_tools.image.renders.previews import Previews


from dataset_tools.image.renders.wrapper import prepare_renders
from dataset_tools.image.renders.separated_anns_grid import SideAnnotationsGrid
from dataset_tools.image.renders.horizontal_grid import HorizontalGrid
from dataset_tools.image.renders.vertical_grid import VerticalGrid
from dataset_tools.image.renders.poster import Poster

from dataset_tools.text.summary.generate import (
    generate_summary_content,
    get_summary_data,
    get_summary_data_sly,
)

from dataset_tools.image.renders.convert import from_mp4_to_webm, compress_mp4, compress_png



import os
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt
import cv2
import math
import supervisely as sly
from skimage.transform import resize
from dataset_tools.image.stats.basestats import BaseVisual


class ClassesHeatmaps(BaseVisual):
    """
    Get heatmaps of visual density of aggregated annotations for every class in the dataset
    """

    def __init__(self, project_meta: sly.ProjectMeta, heatmap_img_size: tuple = None, force=False):

        self.force = force
        self._meta = project_meta
        self.classname_heatmap = {}
        self._ds_image_sizes = []
        self.heatmap_image_paths = []

        if heatmap_img_size:
            self._heatmap_img_size = heatmap_img_size
        else:
            self._heatmap_img_size = (720, 1280)

        for obj_class in self._meta.obj_classes:
            self.classname_heatmap[obj_class.name] = np.zeros(
                self._heatmap_img_size + (3,), dtype=np.float32
            )

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        image_height, image_width = ann.img_size
        self._ds_image_sizes.append((image_height, image_width))
        geometry_types_to_heatmap = ["polygon", "rectangle", "bitmap"]
        ann = ann.resize(self._heatmap_img_size)
        for label in ann.labels:
            temp_canvas = np.zeros(self._heatmap_img_size + (3,), dtype=np.uint8)
            if label.geometry.geometry_name() in geometry_types_to_heatmap:
                label.draw(temp_canvas, color=(1, 1, 1))
                self.classname_heatmap[label.obj_class.name] += temp_canvas

    def to_image(
        self,
        path: str,
        draw_style: str = "inside_white",
        grid_spacing: int = 20,
        outer_grid_spacing: int = 20,
    ) -> None:
        """
        Crates image grid with density heatmaps of all possible classes.

        :param path: path to save output image.
        :type path: str.
        :param draw_style: style in which output heatmaps grid will be represented. Possible values: "inside_white", "outside_black"
        :type draw_style: str, optional
        :param grid_spacing: spaces between images. Defaults to 20.
        :type grid_spacing: int, optional
        :param outer_grid_spacing: frame around the overall image. Defaults to 20.
        :type outer_grid_spacing: int, optional
        """
        self._calculate_output_img_size()

        if draw_style == "inside_white":
            self._create_single_images_text_inside(path)
        if draw_style == "outside_black":
            self._create_single_images_text_outside(path)

        img_height, img_width = cv2.imread(self.heatmap_image_paths[0]).shape[:2]
        num_images = len(self.heatmap_image_paths)
        rows = math.ceil(math.sqrt(num_images))
        cols = math.ceil(num_images / rows)

        result_width = cols * (img_width + grid_spacing) - grid_spacing + 2 * outer_grid_spacing
        result_height = rows * (img_height + grid_spacing) - grid_spacing + 2 * outer_grid_spacing

        result_image = Image.new("RGB", (result_width, result_height), "white")

        for i, img_path in enumerate(self.heatmap_image_paths):
            img = Image.open(img_path)
            row = i // cols
            col = i % cols
            x = outer_grid_spacing + col * (img_width + grid_spacing)
            y = outer_grid_spacing + row * (img_height + grid_spacing)
            result_image.paste(img, (x, y))
            sly.api.file_api.silent_remove(img_path)
            self.heatmap_image_paths = []

        result_image.save(path)

    def _create_single_images_text_outside(self, path):
        for heatmap in self.classname_heatmap:
            resized_image = resize(self.classname_heatmap[heatmap], self._heatmap_img_size)
            image_path = os.path.join(os.path.dirname(path), f"{heatmap}.png")
            plt.imsave(image_path, resized_image[:, :, 0])

            image = cv2.imread(image_path)
            image = self._draw_text_below_image(heatmap, image)
            cv2.imwrite(image_path, image)

            self.heatmap_image_paths.append(image_path)

    def _draw_text_below_image(self, text, image):
        font = cv2.FONT_HERSHEY_SIMPLEX
        line_type = cv2.LINE_AA
        font_scale = 1.0
        text_color = (0, 0, 0)
        thickness = 3
        text_height_percent = 10
        line_spacing_percent = 5

        image_height = image.shape[0]
        text_height = int(image_height * text_height_percent / 100)

        text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
        while text_size[1] > text_height:
            font_scale -= 0.1
            text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
        while text_size[1] < text_height:
            font_scale += 0.1
            text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]

        characters = list(text)
        lines = []
        current_line = characters[0]

        for char in characters[1:]:
            temp_line = current_line + char
            size = cv2.getTextSize(temp_line, font, font_scale, thickness)[0]

            if size[0] <= image.shape[1]:
                current_line = temp_line
            else:
                lines.append(current_line)
                current_line = char

        lines.append(current_line)

        text_x = 10
        text_y = image_height + text_height + 10

        line_spacing = int(image_height * line_spacing_percent / 100)

        result_height = (
            image_height + text_height + (len(lines) - 1) * (text_height + line_spacing) + 40
        )
        result_image = np.zeros((result_height, image.shape[1], 3), dtype=np.uint8)
        result_image.fill(255)
        result_image[:image_height, :] = image

        for line in lines:
            cv2.putText(
                result_image,
                line,
                (text_x, text_y),
                font,
                font_scale,
                text_color,
                thickness,
                line_type,
            )
            text_y += text_height + line_spacing

        return result_image

    def _create_single_images_text_inside(self, path):
        for heatmap in self.classname_heatmap:
            font_scale = self._get_optimal_font_scale(heatmap)
            resized_image = resize(self.classname_heatmap[heatmap], self._heatmap_img_size)
            x_pos_center = int(resized_image.shape[1] * 0.5)
            y_pos_percent = int(resized_image.shape[0] * 0.95)

            image_path = os.path.join(os.path.dirname(path), f"{heatmap}.png")
            plt.imsave(image_path, resized_image[:, :, 0])

            image = cv2.imread(image_path)
            text = f"{heatmap}"
            font = cv2.FONT_HERSHEY_SIMPLEX
            text_color = (255, 255, 255)
            thickness = 3
            line_type = cv2.LINE_AA
            text_width = cv2.getTextSize(text, font, font_scale, thickness)[0][0]
            text_position = (x_pos_center - int(text_width / 2), y_pos_percent)
            cv2.putText(
                image, text, text_position, font, font_scale, text_color, thickness, line_type
            )
            cv2.imwrite(image_path, image)

            self.heatmap_image_paths.append(image_path)

    def _get_optimal_font_scale(self, text):
        font_scale = 10
        thickness = 3
        text_height_percent = 10
        font = cv2.FONT_HERSHEY_SIMPLEX
        text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
        text_width = text_size[0]
        text_height = text_size[1]

        while text_width > self._heatmap_img_size[1]:
            font_scale -= 0.1
            text_size = cv2.getTextSize(text, font, font_scale, thickness)[0]
            text_width = text_size[0]
            text_height = text_size[1]

        desired_text_height = (self._heatmap_img_size[0] * text_height_percent) // 100
        font_scale *= desired_text_height / text_height
        return font_scale

    def _calculate_output_img_size(self):
        sizes = np.array(self._ds_image_sizes)
        if np.all(sizes == sizes[0]):
            self._heatmap_img_size = sizes[0]
        else:
            widths = sizes[:, 1]
            heights = sizes[:, 0]
            median_width = np.median(widths)
            median_height = np.median(heights)
            self._heatmap_img_size = (median_height, median_width)



import os
import random
from typing import List, Union

import supervisely as sly


def sample_images(
    api: sly.Api,
    project: Union[int, str],
    datasets: List[Union[sly.DatasetInfo, sly.Project.DatasetDict]],
    meta: sly.ProjectMeta,
    sample_cnt: float,
):
    total = 0
    samples = []
    for dataset in datasets:
        dataset: sly.Dataset
        k = min(
            sample_cnt // len(datasets),
            (dataset.items_count if isinstance(project, int) else len(os.listdir(dataset.ann_dir))),
        )

        ds_images = (
            api.image.get_list(
                dataset.id,
                filters=[{"field": "labelsCount", "operator": ">", "value": 0}],
            )
            if isinstance(project, int)
            else [
                dataset.get_image_info(sly.fs.get_file_name(img))
                for img in os.listdir(dataset.ann_dir)
                if dataset.get_image_info(sly.fs.get_file_name(img)).labels_count > 0
            ]
        )

        s = random.sample(ds_images, min(k, len(ds_images)))
        anns = (
            [
                sly.Annotation.from_json(ann_json, meta)
                for ann_json in api.annotation.download_json_batch(
                    dataset.id,
                    [item.id for item in s],
                )
            ]
            if isinstance(project, int)
            else [dataset.get_ann(img.name, meta) for img in s]
        )
        samples.append((dataset, s, anns))
        total += k
    return samples, total


def prepare_renders(
    project: Union[int, str], renderers: list, sample_cnt: int = 25, api: sly.Api = None
) -> None:
    if len(renderers) == 0:
        print("Passed 'renderers' parameter is empty. Enable 'force' flag to overwrite renderers output file. Skipping renderers preparation...")
        return
    if api is None:
        api = sly.Api.from_env()
    if isinstance(project, int):
        project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project))
        datasets = api.dataset.get_list(project)
    elif isinstance(project, str):
        project_fs = sly.Project(project, sly.OpenMode.READ)
        project_meta = project_fs.meta
        datasets = project_fs.datasets
    else:
        raise ValueError("Project should be either an integer project ID or a string project path.")

    samples, total = sample_images(api, project, datasets, project_meta, sample_cnt)
    if len(samples) == 0:
        raise Exception("There are not any images with labels on them in the project.")

    for renderer in renderers:
        renderer.update(samples)



import json
import os
from urllib.parse import urljoin

import supervisely as sly


def prepare_download_link(project):
    api = sly.Api.from_env()

    team_id = sly.env.team_id()
    workspace_id = sly.env.workspace_id()
    agent_id = sly.env.agent_id()
    storage_dir = sly.app.get_data_dir()
    local_save_path = os.path.join(storage_dir, "download_links.json")
    api.file.download(team_id, os.environ["DOWNLOADS_DICT"], local_save_path)
    with open(local_save_path, "r") as f:
        links = json.load(f)

    if links.get(str(project.id), None) is not None:
        print("URL already exists. Skipping creation of download link...")
        return links[str(project.id)]
    else:
        print("URL not exists. Creating a download link...")

        app_slug = f"supervisely-ecosystem/export-to-supervisely-format"
        module_id = api.app.get_ecosystem_module_id(app_slug)
        module_info = api.app.get_ecosystem_module_info(module_id)

        print("Start app: ", module_info.name)

        params = module_info.get_arguments(images_project=project.id)

        session = api.app.start(
            agent_id=agent_id,
            module_id=module_id,
            workspace_id=workspace_id,
            task_name="custom session name",
            params=params,
        )
        print("App is started, task_id =", session.task_id)
        print(session)

        try:
            # wait until task end or specific task status
            print("Waiting for the download link to finish being created...")
            api.app.wait(session.task_id, target_status=api.task.Status.FINISHED)

        except sly.WaitingTimeExceeded as e:
            print(e)
            # we don't want to wait more, let's stop our long-lived or "zombie" task
            api.app.stop(session.task_id)
        except sly.TaskFinishedWithError as e:
            print(e)

        # let's list all sessions of specific app in our team with additional optional filtering by statuses [finished]
        sessions = api.app.get_sessions(
            team_id=team_id, module_id=module_id, statuses=[api.task.Status.FINISHED]
        )
        return urljoin(
            os.environ["SERVER_ADDRESS"],
            sessions[0].details["meta"]["output"]["general"]["titleUrl"],
        )


def update_sly_url_dict(new_dict: dict) -> None:
    src = os.environ["DOWNLOADS_DICT"]
    dst = os.path.join(sly.app.get_data_dir(), "download_links.json")

    print("Updating dictionary with download links...")
    api = sly.Api.from_env()
    team_id = sly.env.team_id()

    if api.file.exists(team_id, src):
        api.file.download(team_id, src, dst)

        with open(dst, "r") as f:
            data = json.load(f)
        data.update(new_dict)
    else:
        data = new_dict

    with open(dst, "w") as f:
        json.dump(data, f)

    api.file.upload(team_id, src=dst, dst=src)

    print("Dictionary successfully updated!")



import os
import sys
from collections import defaultdict
from pathlib import Path

import supervisely as sly

# my_app = sly.AppService()
# api: sly.Api = my_app.public_api

root_source_dir = str(Path(sys.argv[0]).parents[1])
sly.logger.info(f"Root source directory: {root_source_dir}")
sys.path.append(root_source_dir)

# TASK_ID = int(os.environ["TASK_ID"])
# TEAM_ID = sly.env.team_id()
# WORKSPACE_ID = sly.env.workspace_id()

logger = sly.logger

# train_ds = os.environ["modal.state.train"]
# test_ds = os.environ["modal.state.test"]

datasets = ["Train", "Test"]

# for ds in [train_ds, test_ds]:
#     if len(ds) != 2:
#         datasets.append(ds[1:-1].replace('\'', ''))

# if len(datasets) == 0:
#     logger.warn('You have not selected a dataset to import')
#     my_app.stop()

train_percent = 100
test_percent = 100

sample_img_count = {"Train": round(2.22 * train_percent), "Test": round(0.55 * test_percent)}

project_name = "TomatoOD"
work_dir = "tomato_data"
images_url = "https://datasets-u2m.s3.eu-west-3.amazonaws.com/tomatOD_images.zip"
annotations_url = "https://datasets-u2m.s3.eu-west-3.amazonaws.com/tomatOD_annotations.zip"

images_arch_name = "tomatOD_images.zip"
annotations_arch_name = "tomatOD_annotations.zip"

anns_folder = "tomatOD_annotations"
ann_prefix = "tomatOD_"
ann_ext = ".json"
batch_size = 30
class_name = "tomato"
train_ds = "Train"

obj_class = sly.ObjClass(class_name, sly.Rectangle)
obj_class_collection = sly.ObjClassCollection([obj_class])

tag_meta_unripe = sly.TagMeta("unripe", sly.TagValueType.NONE)
tag_meta_semi_ripe = sly.TagMeta("semi-ripe", sly.TagValueType.NONE)
tag_meta_fully_ripe = sly.TagMeta("fully-ripe", sly.TagValueType.NONE)

tag_metas = [tag_meta_unripe, tag_meta_semi_ripe, tag_meta_fully_ripe]

tag_meta_collection = sly.TagMetaCollection(tag_metas)
meta = sly.ProjectMeta(obj_classes=obj_class_collection, tag_metas=tag_meta_collection)

storage_dir = sly.app.get_data_dir()
work_dir_path = os.path.join(storage_dir, work_dir)
sly.io.fs.mkdir(work_dir_path)
images_archive_path = os.path.join(work_dir_path, images_arch_name)
annotations_archive_path = os.path.join(work_dir_path, annotations_arch_name)

image_name_to_id = {}
id_to_bbox_anns = defaultdict(list)
id_to_tag = defaultdict(list)
name_to_size = {}
category_id_to_name = {}



import os
import random
from typing import List, Union

import cv2
import numpy as np
from tqdm import tqdm

import supervisely as sly
from dataset_tools.image.renders.convert import compress_png
from supervisely.imaging import font as sly_font


class Poster:
    def __init__(
        self,
        project: Union[str, int],
        project_meta: sly.ProjectMeta,
        api: sly.Api = None,
        force: bool = False,
    ) -> None:
        self.force = force
        self._project_meta = project_meta
        self._api = api if api is not None else sly.Api.from_env()  # ?
        self._project = None
        self._images_info = None
        self._items_count = 0
        self._total_labels = 0
        self._local = False

        if isinstance(project, int):
            self._project = self._api.project.get_info_by_id(project)
            self._items_count = self._project.items_count
            for ds in self._api.dataset.get_list(self._project.id):
                for img in self._api.image.get_list(ds.id):
                    self._total_labels += img.labels_count
        elif isinstance(project, str):
            self._local = True
            self._project = sly.Project(project, sly.OpenMode.READ)
            self._items_count = self._project.total_items
            for ds in self._project.datasets:
                ds: sly.Dataset
                img_infos = [
                    ds.get_image_info(sly.fs.get_file_name(ann_name))
                    for ann_name in os.listdir(ds.ann_dir)
                ]
                for img in img_infos:
                    self._total_labels += img.labels_count

        else:
            raise Exception('Parameter "project" has to be one of `int` or `str` types.')

        self._size = (1080, 1960)
        self._GAP = 20
        self._size_line_1 = (
            (self._size[0] - 3 * self._GAP) // 5 * 3,
            (self._size[1] - 4 * self._GAP) // 3,
        )
        self._size_line_2 = (
            (self._size[0] - 3 * self._GAP) // 5 * 2,
            (self._size[1] - 5 * self._GAP) // 4,
        )

        self._poster = np.ones((*self._size, 4), dtype=np.uint8) * 255  # result poster

        self._logo_path = "logo.png"

    @property
    def basename_stem(self) -> None:
        return sly.utils.camel_to_snake(self.__class__.__name__)

    def update(self, data: tuple):
        np_images = []
        join_data = [(ds, img, ann) for ds, list1, list2 in data for img, ann in zip(list1, list2)]
        random.shuffle(join_data)
        i = 0
        with tqdm(desc="Downloading 7 sample images.", total=7) as pbar:
            while len(np_images) < 7:
                if i > len(join_data) * 3:
                    raise Exception("There not enought images with labels in the project.")
                ds, img_info, ann = join_data[i % len(join_data)]
                ds: sly.Dataset
                if len(ann.labels) < 1:
                    i += 1
                    continue
                np_img = (
                    sly.image.read(ds.get_img_path(img_info.name))
                    if self._local
                    else self._api.image.download_np(img_info.id)
                )

                h, w = np_img.shape[:2]
                background = np.ones((h, w, 3), dtype=np.uint8) * 255
                ann.draw_pretty(np_img, thickness=0, opacity=0.7)
                np_img = cv2.addWeighted(np_img, 0.8, background, 0.2, 0)

                # backup
                # if len(np_images) % 2 == 0:
                #     h, w = np_img.shape[:2]
                #     background = np.ones((h, w, 3), dtype=np.uint8) * 255
                #     alpha = 0.5
                #     np_img = cv2.addWeighted(background, 1 - alpha, np_img, alpha, 0)
                #     ann.draw_pretty(np_img, thickness=0, opacity=0.6)
                # else:
                #     ann: sly.Annotation
                #     thickness = 3
                #     ann.draw_contour(np_img, thickness=thickness)

                np_images.append(self._resize_image(np_img, i))
                i += 1
                pbar.update(1)

        self._create_frame(np_images)
        self._draw_text_and_rectangles()

    def to_image(self, path: str = None):
        path_part, ext = os.path.splitext(path)
        tmp_path = f"{path_part}-o{ext}"
        if path is None:
            storage_dir = sly.app.get_data_dir()
            path = os.path.join(storage_dir, "horizontal_grid.png")

        sly.image.write(tmp_path, self._poster)
        compress_png(tmp_path, path)
        sly.fs.silent_remove(tmp_path)
        sly.logger.info(f"Poster saved to: {path}")

    def _resize_image(self, image: np.ndarray, image_num: int):
        h, w = self._size_line_1 if image_num < 3 else self._size_line_2
        img_h, img_w = image.shape[:2]
        scale_ratio = max(h / img_h, w / img_w)
        img_h, img_w = int(img_h * scale_ratio), int(img_w * scale_ratio)
        image = sly.image.resize(image, (img_h, img_w))
        start_h = (img_h - h) // 2
        start_w = (img_w - w) // 2
        rect = (start_h, start_w, start_h + h, start_w + w)
        image = sly.image.crop_with_padding(image, sly.Rectangle(*rect))

        rgba_image = np.zeros((h, w, 4), dtype=np.uint8)
        rgba_image[:h, :w, :3] = image[:h, :w, :3]
        return rgba_image

    def _draw_text_and_rectangles(self):
        base_font_size = sly_font.get_readable_font_size(self._size)
        font_name_title = "FiraSans-Regular.ttf"
        font_name_subs = "FiraSans-Thin.ttf"
        font_title = sly_font.get_font(font_name_title, int(base_font_size * 6))
        font_subs = sly_font.get_font(font_name_subs, int(base_font_size * 1.8))
        title, *subs = self._get_text_from_project()
        _, _, _, subs_height = font_subs.getbbox("".join(subs))

        poster_h, poster_w = self._size

        bg_image = np.zeros((poster_h, poster_w, 3), dtype=np.uint8)

        title_bottom = self._draw_title(bg_image, font_title, title)
        sub_imgs = [self._draw_subtitles(font_subs, text, subs_height) for text in subs]
        logo = self._draw_logo(subs_height)

        image = np.hstack(sub_imgs)
        subs_h, subs_w = image.shape[:2]
        subs_x = (poster_w - subs_w - logo.shape[1]) // 2
        subs_y = title_bottom + self._GAP

        bg = bg_image[subs_y : subs_y + subs_h, subs_x : subs_x + subs_w, :3]

        alpha_channel = np.where(np.all(image == 0, axis=-1), 0, 255).astype(np.uint8)
        text_image_bw = cv2.cvtColor(image, cv2.COLOR_BGRA2GRAY)
        _, text_mask = cv2.threshold(text_image_bw, 255, 255, cv2.THRESH_BINARY)
        inverse_text_mask = cv2.bitwise_not(text_mask)
        background_masked = cv2.bitwise_and(bg, bg, mask=inverse_text_mask)
        text_image_alpha = cv2.merge(
            (image[:, :, 0], image[:, :, 1], image[:, :, 2], alpha_channel)
        )
        text_image_alpha = cv2.cvtColor(text_image_alpha, cv2.COLOR_BGRA2BGR)
        result = cv2.add(background_masked, text_image_alpha)
        result = np.hstack([result, logo])

        h, w = result.shape[:2]
        cv2.rectangle(result, (0, 0), (w - 1, h - 1), (255, 255, 255), 2)
        self._poster[subs_y : subs_y + h, subs_x : subs_x + w, :3] = result

    def _draw_title(self, image, font, text):
        poster_h, poster_w = self._size
        l, t, r, b = font.getbbox(text)
        title_w, title_h = r - l, b
        while title_w > poster_w * 0.9:
            font = font.font_variant(size=int(font.size * 0.96))
            l, t, r, b = font.getbbox(text)
            title_w, title_h = r - l, b
        title_x, title_y = (poster_w - title_w) // 2, (poster_h - title_h) // 2
        left, top, right, bottom = title_x, title_y, title_x + title_w, title_y + title_h

        self._gradient(image, title_x, 0, title_x + title_w, poster_h)
        p = self._GAP // 3
        image[top + p : bottom - p, left + p : right - p] = 255

        sly.image.draw_text(
            image,
            text,
            (top - t // 2, left),
            font=font,
            fill_background=False,
            color=(0, 0, 0, 210),
        )

        self._poster[top:bottom, left:right, :3] = image[top:bottom, left:right, :3]
        return bottom + p

    def _draw_subtitles(self, font, text, height):
        _, t, r, _ = font.getbbox(text)
        pad = self._GAP // 4
        w = r + self._GAP
        image = np.ones((height, w, 3), dtype=np.uint8) * 255
        sly.image.draw_text(
            image,
            text,
            (-t // 2, pad * 2),
            font=font,
            fill_background=False,
        )
        image[:, r + 3 * pad :, :3] = 0
        image = np.dstack((image, np.ones((*image.shape[:2], 1), dtype=np.uint8) * 255))
        return 255 - image

    def _draw_logo(self, height):
        logo = sly.image.read(self._logo_path)
        logo_h, logo_w = logo.shape[:2]
        scale_factor = height / logo_h
        logo_h, logo_w = height, int(logo_w * scale_factor)

        logo = sly.image.resize(logo, (logo_h, logo_w))
        image = np.zeros((logo_h, logo_w, 4), dtype=np.uint8)
        image[:, :, :3] = logo[:, :, :3]
        return logo

    def _get_text_from_project(self):
        texts = []
        texts.append(
            self._project.name.upper()
            if len(self._project.name.split(" ")) < 4
            else " ".join(self._project.name.split(" ")[:3]).upper()
        )

        texts.append(f"{self._items_count} IMAGES")
        classes_cnt = len(self._project_meta.obj_classes)
        classes_text = f'{classes_cnt} {"CLASS" if classes_cnt == 1 else "CLASSES"}'
        texts.append(classes_text)
        if self._total_labels:
            texts.append(f"{self._total_labels} LABELS")

        return texts

    def _gradient(self, img, left, top, right, bottom):
        c1 = np.array((225, 181, 62))  # rgb
        c2 = np.array((219, 84, 150))  # rgb
        im = np.zeros((bottom - top, right - left, 3), dtype=np.uint8)
        row = np.linspace(0, 1, right - left)
        kernel_1d = np.tile(row, (bottom - top, 1))
        kernel = cv2.merge((kernel_1d, kernel_1d, kernel_1d))
        im = kernel * c1 + (1 - kernel) * c2
        im = im.astype(np.uint8)
        img[top:bottom, left:right, :3] = im
        return img

    def _create_frame(self, images: List[np.ndarray]):
        x, y = self._GAP, self._GAP
        x_end, y_end = x, y

        for img in images[:3]:
            x_end, y_end = x + img.shape[1], y + img.shape[0]
            self._poster[y:y_end, x:x_end] = img
            x = x_end + self._GAP

        y = y_end + self._GAP
        x = self._GAP

        for img in images[3:]:
            x_end, y_end = x + img.shape[1], y + img.shape[0]
            self._poster[y:y_end, x:x_end] = img
            x = x_end + self._GAP



import operator
import os
import re
import textwrap
from typing import Dict, List

import inflect

import supervisely as sly

p = (
    inflect.engine()
)  # correctly generate plurals, singular nouns, ordinals, indefinite articles; convert numbers to words.


def list2sentence(lst: List[str], anytail: str = "", keeptail=False):
    assert isinstance(lst, list) and all(
        isinstance(item, str) for item in lst
    ), "All items in the list must be strings."

    anytail = " " + anytail if anytail != "" else anytail
    if len(lst) == 0:
        raise ValueError("Provided list is empty")

    if len(lst) == 1:
        sentence = lst[0]
    elif len(lst) == 2:
        sentence = " and ".join(lst)
    else:
        sentence = ", ".join(lst[:-1]) + ", and " + lst[-1]

    if keeptail:
        return sentence + anytail

    if anytail != "":
        if p.singular_noun(anytail):
            sentence += anytail if len(lst) > 1 else p.singular_noun(anytail)
        else:
            sentence += p.plural_noun(anytail) if len(lst) > 1 else anytail

    return sentence.strip()


def standardize(text: str):
    return re.sub(r"[_-]", " ", text).strip()


def get_summary_data(
    name: str,
    fullname: str,
    cv_tasks: List[str],
    annotation_types: List[str],
    industries: str,
    release_year: int,
    homepage_url: str,
    license: str,
    license_url: str,
    preview_image_id: int,
    github_url: str,
    download_sly_url: str,
    download_original_url: str = None,
    paper: str = None,
    citation_url: str = None,
    organization_name: str = None,
    organization_url: str = None,
    tags: List[str] = None,
    **kwargs,
) -> Dict:
    api = sly.Api.from_env()
    project_id = kwargs.get("project_id", None)
    project_info = api.project.get_info_by_id(project_id)

    stats = api.project.get_stats(project_id)

    notsorted = [
        [cls["objectClass"]["name"], cls["total"]] for cls in stats["images"]["objectClasses"]
    ]
    totals_dct = {
        "total_assets": stats["images"]["total"]["imagesInDataset"],
        "total_objects": stats["objects"]["total"]["objectsInDataset"],
        "total_classes": len(stats["images"]["objectClasses"]),
        "top_classes": list(
            map(operator.itemgetter(0), sorted(notsorted, key=operator.itemgetter(1), reverse=True))
        ),
    }

    unlabeled_num = stats["images"]["total"]["imagesNotMarked"]
    unlabeled_percent = round(unlabeled_num / totals_dct["total_assets"] * 100)

    splits_list = [
        {"name": item["name"], "split_size": item["imagesCount"]}
        for item in stats["datasets"]["items"]
    ]

    fields = {
        # preset fields
        "name": name,
        "fullname": fullname,
        "cv_tasks": cv_tasks,
        "annotation_types": annotation_types,
        "industries": industries,
        "release_year": release_year,
        "homepage_url": homepage_url,
        "license": license,
        "license_url": license_url,
        "preview_image_id": preview_image_id,
        "github_url": github_url,
        "citation_url": citation_url,
        "download_sly_url": download_sly_url,
        # from supervisely
        "modality": project_info.type,
        "totals": totals_dct,
        "unlabeled_assets_num": unlabeled_num,
        "unlabeled_assets_percent": unlabeled_percent,
        "splits": splits_list,
    }

    # optional fields
    for key, value in zip(
        ["download_original_url", "paper", "organization_name", "organization_url", "tags"],
        [download_original_url, paper, organization_name, organization_url, tags],
    ):
        if value is not None:
            fields[key] = value

    return fields


def generate_summary_content(data: Dict, vis_url: str) -> str:
    # preset fields
    # required
    name = data.get("name", None)
    fullname = data.get("fullname", None)
    cv_tasks = [standardize(cv_task) for cv_task in data.get("cv_tasks", [])]
    annotation_types = [standardize(ann_type) for ann_type in data.get("annotation_types", [])]
    industries = data.get("industries", None)
    release_year = data.get("release_year", None)
    homepage_url = data.get("homepage_url", None)
    license = data.get("license", None)
    license_url = data.get("license_url", None)
    preview_image_id = data.get("preview_image_id", None)
    github_url = data.get("github_url", None)
    citation_url = data.get("citation_url", None)
    download_sly_url = data.get("download_sly_url", None)

    # optional
    download_original_url = data.get("download_original_url", None)
    paper = data.get("paper", None)
    organization_name = data.get("organization_name", None)
    organization_url = data.get("organization_url", None)
    tags = (data.get("tags", []),)

    # from supervisely
    # required
    modality = data.get("modality")
    totals = data.get("totals", {})
    top_classes = totals.get("top_classes", [])
    unlabeled_assets_num = data.get("unlabeled_assets_num", None)
    unlabeled_assets_percent = data.get("unlabeled_assets_percent", None)
    splits = [
        f'*{split["name"]}* ({split["split_size"]} {modality})' for split in data.get("splits", [])
    ]

    # prepare data
    annotations = []
    if "instance segmentation" in annotation_types:
        if (
            "semantic segmentation" not in annotation_types
            and "object detection" not in annotation_types
        ):
            annotations.append(
                " pixel-level instance segmentation annotations. Due to the nature of the instance segmentation task, it can be automatically transformed into semantic segmentation (only one mask for every class) or object detection (bounding boxes for every object) tasks"
            )
        elif (
            "semantic segmentation" in annotation_types
            and "object detection" not in annotation_types
        ):
            annotations.append(
                " pixel-level instance segmentation annotations. Due to the nature of the instance segmentation task, it can be automatically transformed into object detection (bounding boxes for every object) task"
            )
        elif (
            "semantic segmentation" not in annotation_types
            and "object detection" in annotation_types
        ):
            annotations.append(
                " pixel-level instance segmentation annotations. Due to the nature of the instance segmentation task, it can be automatically transformed into semantic segmentation task (only one mask for every class)"
            )
        else:
            annotations.append(" pixel-level instance segmentation annotations")
    else:
        if "semantic segmentation" in annotation_types:
            annotations.append(" pixel-level semantic segmentation annotations")
        if "object detection" in annotation_types:
            annotations.append(" bounding boxes annotations")

    annotations = ",".join(annotations).strip()

    # collect content
    content = f"**{name}** ({fullname}) is a dataset for {list2sentence(cv_tasks, 'tasks', keeptail=True)}. "

    if "general domain" in industries:
        content += "It is applicable or relevant across various domains."
        if len(industries) > 1:
            industries.pop("general domain")
            content += f"Also, it is used in {list2sentence(industries, 'industries')}."
    else:
        content += f"It is used in the {list2sentence(industries, 'industries')}."

    content += "\n\n"
    content += f"The dataset consists of {totals.get('total_assets', 0)} {modality} with {totals.get('total_objects', 0)} labeled objects belonging to {totals.get('total_classes', 0)} "
    if len(top_classes) == 1:
        content += f"single class "
    else:
        content += f"different classes "
    if len(top_classes) > 3:
        content += f"including *{'*, *'.join(top_classes[:3])},* and other: *{list2sentence(top_classes[3:], )}*."
    elif len(top_classes) == 1:
        content += f"(*{top_classes[0]}*)."
    else:
        content += f"including *{list2sentence(top_classes[:3])}*."
    content += f"\n\nEach {p.singular_noun(modality)} in the {name} dataset has {annotations}. "
    content += f"There are {unlabeled_assets_num} ({unlabeled_assets_percent}% of the total) unlabeled {modality} (i.e. without annotations). "
    content += f"There are {len(splits)} splits in the dataset: {list2sentence(splits)}. "
    if organization_name is not None and organization_url is not None:
        content += f"The dataset was released in {release_year} by the [{organization_name}]({organization_url})."
    elif organization_name is not None and organization_url is None:
        content += f"The dataset was released in {release_year} by the {organization_name}."
    elif organization_name is None and organization_url is None:
        content += f"The dataset was released in {release_year}."

    if len(top_classes) == 1:
        content += f"\n\nHere is the visualized example of the single {top_classes[0]} class:\n\n"
    else:
        content += f"\n\nHere are the visualized examples for each of the {totals.get('total_classes', 0)} classes:\n\n"
    content += f"[Dataset classes]({vis_url})\n"

    return content


def get_summary_data_sly(project_info: sly.ProjectInfo) -> Dict:
    return get_summary_data(**project_info.custom_data, project_id=project_info.id)


# def generate_meta_from_local():

#     modality ="images"

#     with open("./stats/class_balance.json") as f:
#         json_data = json.load(f)
#     df = pd.DataFrame(data=json_data["data"], columns=json_data["columns"])

#     with open("./stats/classes_per_image.json") as f:
#         json_data = json.load(f)
#     df_img = pd.DataFrame(data=json_data["data"], columns=json_data["columns"])


#     totals_dct = {
#        "total_modality_files": df_img.shape[0],
#        "total_objects": df["Objects"].sum(),
#        "total_classes": df["Class"].count(),
#        "top_classes": df.sort_values('Objects', ascending=False)['Class'].tolist()
#     }

#     unlabeled_num = df_img.shape[0] - sum(df_img.drop(columns=["Image","Split", "Height", "Width", "Unlabeled"]).sum(axis=1)==0)
#     unlabeled_percent = unlabeled_num / df_img.shape[0]
#     splits_list = [
#         {
#         "name": "training",
#         "split_size": 800
#         },
#         {
#         "name": "validation",
#         "split_size": 200
#         }
#     ]

#     return {
#         "name": "PASCAL VOC",
#         "fullname": "PASCAL Visual Object Classes Challenge",
#         "cv_tasks": ["semantic-segmentation"],
#         "modality": modality,
#         "release_year": "2012",
#         "organization": "Dong et al",
#         "organization_link": "https://arxiv.org/pdf/2012.07131v2.pdf",
#         "totals": totals_dct,
#         "unlabeled_assets_num": unlabeled_num,
#         "unlabeled_assets_percent": unlabeled_percent,
#         "splits": splits_list
#     }



import os
import random
import zipfile
from typing import Literal

import gdown
from tqdm import tqdm

import dataset_tools.convert.tomatod.sly_globals as g
import supervisely as sly
from supervisely.io.fs import get_file_name
from supervisely.io.json import load_json_file


def prepare_ann_data(ann_path):
    ann_json = load_json_file(ann_path)
    annotations = ann_json["annotations"]
    images = ann_json["images"]

    for image_data in images:
        g.image_name_to_id[image_data["file_name"]] = image_data["id"]
        g.name_to_size[image_data["file_name"]] = (image_data["height"], image_data["width"])

    for ann_data in annotations:
        g.id_to_bbox_anns[ann_data["image_id"]].append(ann_data["bbox"])
        g.id_to_tag[ann_data["image_id"]].append(ann_data["category_id"])

    for category in ann_json["categories"]:
        g.category_id_to_name[category["id"]] = category["name"]


def create_ann(img_name, ds_name):
    labels = []

    im_id = g.image_name_to_id[img_name]
    img_size = g.name_to_size[img_name]
    bbox_anns = g.id_to_bbox_anns[im_id]
    tag_ids = g.id_to_tag[im_id]

    for idx, bbox in enumerate(bbox_anns):
        rectangle = sly.Rectangle(bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2])

        if ds_name == g.train_ds:
            tag_name = g.category_id_to_name[tag_ids[idx]]
        else:  # test_ann diff from train_ann in tags data
            tag_name = g.category_id_to_name[1][tag_ids[idx] - 1]
        tag = sly.Tag(g.meta.get_tag_meta(tag_name))

        label = sly.Label(rectangle, g.obj_class, tags=sly.TagCollection([tag]))
        labels.append(label)

    return sly.Annotation(img_size=img_size, labels=labels)


def extract_zip(archive_path):
    if zipfile.is_zipfile(archive_path):
        with zipfile.ZipFile(archive_path, "r") as archive:
            archive.extractall(g.work_dir_path)
    else:
        g.logger.warn("Archive cannot be unpacked {}".format(get_file_name(archive_path)))
        g.my_app.stop()


def to_supervisely(api, WORKSPACE_ID):
    if not os.path.exists(g.images_archive_path):
        gdown.download(g.images_url, g.images_archive_path, quiet=False)
    extract_zip(g.images_archive_path)
    if not os.path.exists(g.annotations_archive_path):
        gdown.download(g.annotations_url, g.annotations_archive_path, quiet=False)
    extract_zip(g.annotations_archive_path)

    anns_path = os.path.join(g.work_dir_path, g.anns_folder)

    new_project = api.project.create(WORKSPACE_ID, g.project_name, change_name_if_conflict=True)
    api.project.update_meta(new_project.id, g.meta.to_json())

    for ds in g.datasets:
        new_dataset = api.dataset.create(new_project.id, ds, change_name_if_conflict=True)

        curr_img_path = os.path.join(g.work_dir_path, ds.lower())
        curr_ann_path = os.path.join(anns_path, g.ann_prefix + ds.lower() + g.ann_ext)
        prepare_ann_data(curr_ann_path)

        curr_img_cnt = g.sample_img_count[ds]
        sample_img_path = random.sample(os.listdir(curr_img_path), curr_img_cnt)

        with tqdm(desc="Create dataset {}".format(ds), total=curr_img_cnt) as pbar:
            for img_batch in sly.batched(sample_img_path, batch_size=g.batch_size):
                img_pathes = [os.path.join(curr_img_path, name) for name in img_batch]
                img_infos = api.image.upload_paths(new_dataset.id, img_batch, img_pathes)
                img_ids = [im_info.id for im_info in img_infos]

                anns = [create_ann(img_name, ds) for img_name in img_batch]
                api.annotation.upload_anns(img_ids, anns)

                pbar.update(len(img_batch))

    return new_project.id


def from_supervisely(
    input_path: str, output_path: str = None, to_format: Literal["dir", "tar", "both"] = "both"
) -> str:
    pass



import json
import os
import shutil
from datetime import datetime
from typing import List

import numpy as np
import pycocotools.mask as mask_util
import requests
from dotenv import load_dotenv
from PIL import Image
from pycocotools.coco import COCO
from tqdm import tqdm

import supervisely as sly
from dataset_tools.convert import unpack_if_archive
from supervisely.annotation.json_geometries_map import GET_GEOMETRY_FROM_STR


images_links = {
    "train2014": "http://images.cocodataset.org/zips/train2014.zip",
    "val2014": "http://images.cocodataset.org/zips/val2014.zip",
    "test2014": "http://images.cocodataset.org/zips/test2014.zip",
    "train2017": "http://images.cocodataset.org/zips/train2017.zip",
    "val2017": "http://images.cocodataset.org/zips/val2017.zip",
    "test2017": "http://images.cocodataset.org/zips/test2017.zip",
}

annotations_links = {
    "trainval2014": "http://images.cocodataset.org/annotations/annotations_trainval2014.zip",
    "trainval2017": "http://images.cocodataset.org/annotations/annotations_trainval2017.zip",
}

data_dir = sly.app.get_data_dir()

COCO_BASE_DIR = os.path.join(data_dir, "COCO")
META = sly.ProjectMeta()
img_dir = None
ann_dir = None
src_img_dir = None
dst_img_dir = None


def to_supervisely(
    original_ds_names: List[str] = None,
    custom_ds_paths: List[str] = None,
    dst_path: str = None,
):
    global img_dir, ann_dir, src_img_dir, dst_img_dir, META
    if original_ds_names is not None and custom_ds_paths is not None:
        raise ValueError("Both original and custom arguments are given, but only one is allowed.")
    is_original = True if original_ds_names is not None else False
    dst_path = os.path.join(data_dir, "supervisely project") if dst_path is None else dst_path

    if is_original:
        coco_datasets_dirs = download_original_coco_dataset(original_ds_names)
        coco_datasets_name = original_ds_names
    else:
        coco_datasets_dirs = custom_ds_paths
        coco_datasets_name = [os.path.basename(ds) for ds in custom_ds_paths]

    for dataset_name, coco_dataset_dir in zip(coco_datasets_name, coco_datasets_dirs):
        if not sly.fs.dir_exists(coco_dataset_dir):
            sly.logger.info(f"File {coco_dataset_dir} has been skipped.")
            continue
        coco_ann_dir = os.path.join(coco_dataset_dir, "annotations")
        if not sly.fs.dir_exists(os.path.join(coco_dataset_dir, "images")):
            sly.logger.warn(
                "Incorrect input data. Folder with images must be named 'images'. See 'README' for more information."
            )
            continue

        coco_ann_path = get_ann_path(coco_ann_dir, dataset_name, is_original)
        if bool(os.path.exists(coco_ann_path) and os.path.isfile(coco_ann_path)):
            coco = COCO(annotation_file=coco_ann_path)
            categories = coco.loadCats(ids=coco.getCatIds())
            coco_images = coco.imgs
            coco_anns = coco.imgToAnns

            sly_dataset_dir = create_sly_dataset_dir(dst_path, dataset_name=dataset_name)
            img_dir = os.path.join(sly_dataset_dir, "img")
            ann_dir = os.path.join(sly_dataset_dir, "ann")
            meta = get_sly_meta_from_coco(META, dst_path, categories, dataset_name)

            ds_progress = tqdm(desc=f"Converting dataset: {dataset_name}", total=len(coco_images))

            for img_id, img_info in coco_images.items():
                img_ann = coco_anns[img_id]
                img_size = (img_info["height"], img_info["width"])
                ann = create_sly_ann_from_coco_annotation(
                    meta=meta,
                    coco_categories=categories,
                    coco_ann=img_ann,
                    image_size=img_size,
                )
                move_trainvalds_to_sly_dataset(
                    dataset_dir=coco_dataset_dir, coco_image=img_info, ann=ann
                )
                ds_progress.update(1)
        else:
            sly_dataset_dir = create_sly_dataset_dir(dst_path, dataset_name=dataset_name)
            src_img_dir = os.path.join(coco_dataset_dir, "images")
            dst_img_dir = os.path.join(sly_dataset_dir, "img")
            ann_dir = os.path.join(sly_dataset_dir, "ann")
            move_testds_to_sly_dataset(dataset=dataset_name)

    if is_original:
        sly.fs.remove_dir(COCO_BASE_DIR)
    sly.logger.info(f"COCO dataset converted to Supervisely project: {dst_path}")
    return dst_path


def from_supervisely(
    src_path: str,
    dst_path: str = None,
    only_annotated_images: bool = True,
    only_annotations: bool = True,
):
    src_path = unpack_if_archive(src_path)
    parent_dir = os.path.dirname(os.path.normpath(src_path))
    coco_base_dir = os.path.join(parent_dir, "coco project") if dst_path is None else dst_path
    project = sly.Project(src_path, sly.OpenMode.READ)
    meta = project.meta
    datasets = project.datasets

    meta = prepare_meta(meta)
    categories_mapping = get_categories_map_from_meta(meta)
    sly.fs.mkdir(coco_base_dir)
    label_id = 0

    for dataset in datasets:
        sly.logger.info(f"processing {dataset.name}...")
        coco_dataset_dir = os.path.join(coco_base_dir, dataset.name)
        img_dir, ann_dir = create_coco_dataset(coco_dataset_dir)

        coco_ann = {}
        images = os.listdir(dataset.img_dir)

        ds_progress = tqdm(desc=f"Converting dataset: {dataset.name}", total=len(images))
        for batch in sly.batched(images):
            tmp_anns = [dataset.get_ann(name, meta) for name in batch]
            image_names = []
            anns = []
            if only_annotated_images is True:
                for ann, img_name in zip(tmp_anns, batch):
                    if len(ann.labels) > 0:
                        anns.append(ann)
                        image_names.append(img_name)
            else:
                image_names.extend(batch)
                anns.extend(tmp_anns)

            if only_annotations is False:
                src_paths = [dataset.get_img_path(name) for name in image_names]
                dst_paths = [os.path.join(img_dir, name) for name in image_names]
                for src_path, dst in zip(src_paths, dst_paths):
                    shutil.copyfile(src_path, dst)

            anns = [convert_annotation(ann, meta) for ann in anns]
            user_name = "Supervisely"
            coco_ann, label_id = create_coco_annotation(
                meta,
                categories_mapping,
                dataset,
                user_name,
                image_names,
                anns,
                label_id,
                coco_ann,
                ds_progress,
            )
        with open(os.path.join(ann_dir, "instances.json"), "w") as file:
            json.dump(coco_ann, file)

        sly.logger.info(f"dataset {dataset.name} processed!")

    return coco_base_dir


def download_original_coco_dataset(datasets):
    datasets_dirs = []
    for dataset in datasets:
        dataset_dir = os.path.join(COCO_BASE_DIR, dataset)
        sly.fs.mkdir(dataset_dir)
        archive_path = f"{dataset_dir}.zip"
        download_coco_images(dataset, archive_path, dataset_dir)
        if not dataset.startswith("test"):
            download_coco_annotations(dataset, archive_path, dataset_dir)
        datasets_dirs.append(dataset_dir)
    return datasets_dirs


def download_coco_images(dataset, archive_path, save_path):
    link = images_links[dataset]
    response = requests.head(link, allow_redirects=True)
    sizeb = int(response.headers.get("content-length", 0))
    p = tqdm(desc=f"Downloading COCO images", total=sizeb, unit="B", unit_scale=True)
    if not sly.fs.file_exists(archive_path):
        sly.fs.download(link, archive_path, progress=p.update)
    shutil.unpack_archive(archive_path, save_path, format="zip")
    os.rename(os.path.join(save_path, dataset), os.path.join(save_path, "images"))
    sly.fs.silent_remove(archive_path)


def download_coco_annotations(dataset, archive_path, save_path):
    link = None
    ann_dir = os.path.join(save_path, "annotations")
    if dataset in ["train2014", "val2014"]:
        if os.path.exists(ann_dir):
            return
        link = annotations_links["trainval2014"]
    elif dataset in ["train2017", "val2017"]:
        if os.path.exists(ann_dir):
            return
        link = annotations_links["trainval2017"]
    response = requests.head(link, allow_redirects=True)
    sizeb = int(response.headers.get("content-length", 0))
    p = tqdm(desc=f"Downloading COCO anns", total=sizeb, unit="B", unit_scale=True)
    sly.fs.download(link, archive_path, progress=p.update)
    shutil.unpack_archive(archive_path, save_path, format="zip")
    for file in os.listdir(ann_dir):
        if file != f"instances_{dataset}.json":
            sly.fs.silent_remove(os.path.join(ann_dir, file))
    sly.fs.silent_remove(archive_path)


def get_ann_path(ann_dir, dataset_name, is_original):
    if is_original:
        return os.path.join(ann_dir, f"instances_{dataset_name}.json")
    else:
        return os.path.join(ann_dir, "instances.json")


def create_sly_dataset_dir(dst_path, dataset_name):
    dataset_dir = os.path.join(dst_path, dataset_name)
    sly.fs.mkdir(dataset_dir)
    img_dir = os.path.join(dataset_dir, "img")
    sly.fs.mkdir(img_dir)
    ann_dir = os.path.join(dataset_dir, "ann")
    sly.fs.mkdir(ann_dir)
    return dataset_dir


def get_sly_meta_from_coco(meta, dst_path, coco_categories, dataset_name):
    path_to_meta = os.path.join(dst_path, "meta.json")
    if not os.path.exists(path_to_meta):
        meta = dump_meta(meta, coco_categories, path_to_meta)
    elif dataset_name not in ["train2014", "val2014", "train2017", "val2017"]:
        meta = dump_meta(meta, coco_categories, path_to_meta)
    return meta


def dump_meta(meta, coco_categories, path_to_meta):
    meta = create_sly_meta_from_coco_categories(meta, coco_categories)
    meta_json = meta.to_json()
    sly.json.dump_json_file(meta_json, path_to_meta)
    return meta


def create_sly_meta_from_coco_categories(meta, coco_categories):
    colors = []
    for category in coco_categories:
        if category["name"] in [obj_class.name for obj_class in meta.obj_classes]:
            continue
        new_color = sly.color.generate_rgb(colors)
        colors.append(new_color)
        obj_class = sly.ObjClass(category["name"], sly.AnyGeometry, new_color)
        meta = meta.add_obj_class(obj_class)
    return meta


def create_sly_ann_from_coco_annotation(meta, coco_categories, coco_ann, image_size):
    labels = []
    for object in coco_ann:
        name_cat_id_map = coco_category_to_class_name(coco_categories)
        obj_class_name = name_cat_id_map[object["category_id"]]
        obj_class = meta.get_obj_class(obj_class_name)
        if type(object["segmentation"]) is dict:
            polygons = convert_rle_mask_to_polygon(object)
            for polygon in polygons:
                figure = polygon
                label = sly.Label(figure, obj_class)
                labels.append(label)
        elif type(object["segmentation"]) is list and object["segmentation"]:
            figure = convert_polygon_vertices(object)
            label = sly.Label(figure, obj_class)
            labels.append(label)

        bbox = object["bbox"]
        xmin = bbox[0]
        ymin = bbox[1]
        xmax = xmin + bbox[2]
        ymax = ymin + bbox[3]
        rectangle = sly.Label(
            sly.Rectangle(top=ymin, left=xmin, bottom=ymax, right=xmax), obj_class
        )
        labels.append(rectangle)
    return sly.Annotation(image_size, labels)


def coco_category_to_class_name(coco_categories):
    return {category["id"]: category["name"] for category in coco_categories}


def convert_rle_mask_to_polygon(coco_ann):
    if type(coco_ann["segmentation"]["counts"]) is str:
        coco_ann["segmentation"]["counts"] = bytes(
            coco_ann["segmentation"]["counts"], encoding="utf-8"
        )
        mask = mask_util.decode(coco_ann["segmentation"])
    else:
        rle_obj = mask_util.frPyObjects(
            coco_ann["segmentation"],
            coco_ann["segmentation"]["size"][0],
            coco_ann["segmentation"]["size"][1],
        )
        mask = mask_util.decode(rle_obj)
    mask = np.array(mask, dtype=bool)
    return sly.Bitmap(mask).to_contours()


def convert_polygon_vertices(coco_ann):
    for polygons in coco_ann["segmentation"]:
        exterior = polygons
        exterior = [exterior[i * 2 : (i + 1) * 2] for i in range((len(exterior) + 2 - 1) // 2)]
        exterior = [sly.PointLocation(height, width) for width, height in exterior]
        return sly.Polygon(exterior, [])


def move_trainvalds_to_sly_dataset(dataset_dir, coco_image, ann):
    image_name = coco_image["file_name"]
    ann_json = ann.to_json()
    sly.json.dump_json_file(ann_json, os.path.join(ann_dir, f"{image_name}.json"))
    coco_img_path = os.path.join(dataset_dir, "images", image_name)
    sly_img_path = os.path.join(img_dir, image_name)
    if sly.fs.file_exists(os.path.join(coco_img_path)):
        shutil.copy(coco_img_path, sly_img_path)


def move_testds_to_sly_dataset(dataset):
    ds_progress = tqdm(f"Converting dataset: {dataset}", len(os.listdir(src_img_dir)))
    for image in os.listdir(src_img_dir):
        src_image_path = os.path.join(src_img_dir, image)
        dst_image_path = os.path.join(dst_img_dir, image)
        shutil.copy(src_image_path, dst_image_path)
        im = Image.open(dst_image_path)
        width, height = im.size
        img_size = (height, width)
        ann = sly.Annotation(img_size)
        ann_json = ann.to_json()
        sly.json.dump_json_file(ann_json, os.path.join(ann_dir, f"{image}.json"))
        ds_progress.update(1)


def prepare_meta(meta):
    new_classes = []
    for obj_cls in meta.obj_classes:
        obj_cls: sly.ObjClass
        new_classes.append(obj_cls.clone(geometry_type=GET_GEOMETRY_FROM_STR("polygon")))

    meta = meta.clone(obj_classes=sly.ObjClassCollection(new_classes))
    return meta


def get_categories_map_from_meta(meta):
    obj_classes = meta.obj_classes
    categories_mapping = {}
    for idx, obj_class in enumerate(obj_classes):
        categories_mapping[obj_class.name] = idx + 1
    return categories_mapping


def create_coco_dataset(coco_dataset_dir):
    sly.fs.mkdir(os.path.join(coco_dataset_dir))
    img_dir = os.path.join(coco_dataset_dir, "images")
    sly.fs.mkdir(img_dir)
    ann_dir = os.path.join(coco_dataset_dir, "annotations")
    sly.fs.mkdir(ann_dir)
    return img_dir, ann_dir


def convert_annotation(ann: sly.Annotation, dst_meta):
    new_labels = []
    for lbl in ann.labels:
        new_cls = dst_meta.obj_classes.get(lbl.obj_class.name)
        if lbl.obj_class.geometry_type == new_cls.geometry_type:
            new_labels.append(lbl)
        else:
            converted_labels = lbl.convert(new_cls)
            new_labels.extend(converted_labels)
    return ann.clone(labels=new_labels)


def create_coco_annotation(
    meta,
    categories_mapping,
    dataset,
    user_name,
    image_names,
    anns,
    label_id,
    coco_ann,
    progress,
):
    date_created = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    if len(coco_ann) == 0:
        coco_ann = dict(
            info=dict(
                description=dataset.name,
                url="None",
                version=str(1.0),
                year=dataset.name[-4:],
                contributor=user_name,
                date_created=date_created,
            ),
            licenses=[dict(url="None", id=0, name="None")],
            images=[
                # license, url, file_name, height, width, date_captured, id
            ],
            # type="instances",
            annotations=[
                # segmentation, area, iscrowd, image_id, bbox, category_id, id
            ],
            categories=get_categories_from_meta(meta),  # supercategory, id, name
        )

    for image_name, ann in zip(image_names, anns):
        ann: sly.Annotation
        height, width = ann.img_size
        coco_ann["images"].append(
            dict(
                license="None",
                file_name=image_name,
                url="None",  # image_info.full_storage_url,  # coco_url, flickr_url
                height=height,
                width=width,
                date_captured=date_created,
                id=image_name,
            )
        )

        for label in ann.labels:
            segmentation = label.geometry.to_json()["points"]["exterior"]
            segmentation = coco_segmentation(segmentation)

            bbox = label.geometry.to_bbox().to_json()["points"]["exterior"]
            bbox = coco_bbox(bbox)

            label_id += 1
            coco_ann["annotations"].append(
                dict(
                    segmentation=[
                        segmentation
                    ],  # a list of polygon vertices around the object, but can also be a run-length-encoded (RLE) bit mask
                    area=label.geometry.area,  # Area is measured in pixels (e.g. a 10px by 20px box would have an area of 200)
                    iscrowd=0,  # Is Crowd specifies whether the segmentation is for a single object or for a group/cluster of objects
                    image_id=image_name,  # The image id corresponds to a specific image in the dataset
                    bbox=bbox,  # he COCO bounding box format is [top left x position, top left y position, width, height]
                    category_id=categories_mapping[
                        label.obj_class.name
                    ],  # The category id corresponds to a single category specified in the categories section
                    id=label_id,  # Each annotation also has an id (unique to all other annotations in the dataset)
                )
            )
        progress.update(1)
    return coco_ann, label_id


def coco_segmentation(segmentation):  # works only with external vertices for now
    segmentation = [float(coord) for sublist in segmentation for coord in sublist]
    return segmentation


def coco_bbox(bbox):
    bbox = [float(coord) for sublist in bbox for coord in sublist]
    x, y, max_x, max_y = bbox
    width = max_x - x
    height = max_y - y
    bbox = (x, y, width, height)
    return bbox


def get_categories_from_meta(meta):
    obj_classes = meta.obj_classes
    categories = []
    for idx, obj_class in enumerate(obj_classes):
        categories.append(
            dict(
                supercategory=obj_class.name,
                id=idx + 1,  # supercategory id
                name=obj_class.name,
            )
        )
    return categories



from typing import List, Dict

import numpy as np
import supervisely as sly

from dataset_tools.image.stats.basestats import BaseStats

UNLABELED_COLOR = [0, 0, 0]


class ClassesPerImage(BaseStats):
    """
    Columns:
        Image
        Dataset
        Height
        Width
        # Channels
        Unlabeled
        Class1 objects count
        Class1 covered area (%)
        Class2 objects count
        Class2 covered area (%)
        etc.
    """

    def __init__(
        self, project_meta: sly.ProjectMeta, datasets: List[sly.DatasetInfo] = None, force:bool = False
    ) -> None:
        self._meta = project_meta
        self._stats = {}
        self.force = force
        
        self._dataset_id_to_name = None
        if datasets is not None:
            self._dataset_id_to_name = {ds.id: ds.name for ds in datasets}

        self._class_names = ["unlabeled"]
        self._class_indices_colors = [UNLABELED_COLOR]
        self._classname_to_index = {}

        for idx, obj_class in enumerate(self._meta.obj_classes):
            self._class_names.append(obj_class.name)
            class_index = idx + 1
            self._class_indices_colors.append([class_index, class_index, class_index])
            self._classname_to_index[obj_class.name] = class_index

        self._stats["data"] = []
        self._referencesRow = []

    def update(self, image_info: sly.ImageInfo, ann: sly.Annotation) -> None:
        render_idx_rgb = np.zeros(ann.img_size + (3,), dtype=np.uint8)
        render_idx_rgb[:] = UNLABELED_COLOR
        ann.draw_class_idx_rgb(render_idx_rgb, self._classname_to_index)
        stat_area = sly.Annotation.stat_area(
            render_idx_rgb, self._class_names, self._class_indices_colors
        )
        stat_count = ann.stat_class_count(self._class_names)

        if stat_area["unlabeled"] > 0:
            stat_count["unlabeled"] = 1

        table_row = []

        table_row.append(image_info.name)

        if self._dataset_id_to_name is not None:
            table_row.append(self._dataset_id_to_name[image_info.dataset_id])
        area_unl = stat_area["unlabeled"] if not np.isnan(stat_area["unlabeled"]) else 0
        table_row.extend(
            [
                stat_area["height"],
                stat_area["width"],
                # stat_area["channels"],
                round(area_unl, 2) if area_unl != 0 else 0,
            ]
        )
        for class_name in self._class_names:
            cur_area = stat_area[class_name] if not np.isnan(stat_area[class_name]) else 0
            cur_count = stat_count[class_name] if not np.isnan(stat_count[class_name]) else 0
            if class_name == "unlabeled":
                continue
            table_row.append(cur_count)
            table_row.append(round(cur_area, 2) if cur_area != 0 else 0)

        self._stats["data"].append(table_row)
        self._referencesRow.append([image_info.id])

    def to_json(self) -> Dict:
        if self._dataset_id_to_name is not None:
            columns = ["Image", "Split", "Height", "Width", "Unlabeled"]
        else:
            columns = ["Image", "Height", "Width", "Unlabeled"]

        columns_options = [None] * len(columns)

        if self._dataset_id_to_name is not None:
            columns_options[columns.index("Split")] = {
                "subtitle": "folder name",
            }
        columns_options[columns.index("Height")] = {
            "postfix": "px",
        }
        columns_options[columns.index("Width")] = {
            "postfix": "px",
        }
        columns_options[columns.index("Unlabeled")] = {
            "subtitle": "area",
            "postfix": "%",
        }

        for obj_class in self._meta.obj_classes:
            columns_options.append({"subtitle": "objects count"})
            columns_options.append({"subtitle": "covered area", "postfix": "%"})
            columns.extend([obj_class.name] * 2)

        options = {"fixColumns": 1}
        res = {
            "columns": columns,
            "columnsOptions": columns_options,
            "data": self._stats["data"],
            "options": options,
            "referencesRow": self._referencesRow,
        }
        return res



import os
import sys
from collections import defaultdict
from pathlib import Path

import supervisely as sly

# my_app = sly.AppService()
# api: sly.Api = my_app.public_api

root_source_dir = str(Path(sys.argv[0]).parents[1])
sly.logger.info(f"Root source directory: {root_source_dir}")
sys.path.append(root_source_dir)

# TASK_ID = int(os.environ["TASK_ID"])
TEAM_ID = sly.env.team_id()
WORKSPACE_ID = sly.env.workspace_id()

logger = sly.logger

# train_ds = os.environ["modal.state.train"]
# test_ds = os.environ["modal.state.test"]

datasets = ["Train", "Test"]

# for ds in [train_ds, test_ds]:
#     if len(ds) != 2:
#         datasets.append(ds[1:-1].replace('\'', ''))

# if len(datasets) == 0:
#     logger.warn('You have not selected a dataset to import')
#     my_app.stop()

train_percent = 100
test_percent = 100

sample_img_count = {"Train": round(6.43 * train_percent), "Test": round(1.61 * test_percent)}

project_name = "LaboroTomato"
work_dir = "tomato_data"
apple_url = "http://assets.laboro.ai/laborotomato/laboro_tomato.zip"

arch_name = "laboro_tomato.zip"

folder_name = "laboro_tomato"
anns_folder = "annotations"
batch_size = 30
class_name = "tomato"

obj_class = sly.ObjClass(class_name, sly.Polygon)
obj_class_collection = sly.ObjClassCollection([obj_class])

tag_meta_b_fully_ripened = sly.TagMeta("b_fully_ripened", sly.TagValueType.NONE)
tag_meta_b_half_ripened = sly.TagMeta("b_half_ripened", sly.TagValueType.NONE)
tag_meta_b_green = sly.TagMeta("b_green", sly.TagValueType.NONE)
tag_meta_l_fully_ripened = sly.TagMeta("l_fully_ripened", sly.TagValueType.NONE)
tag_meta_l_half_ripened = sly.TagMeta("l_half_ripened", sly.TagValueType.NONE)
tag_meta_l_green = sly.TagMeta("l_green", sly.TagValueType.NONE)

tag_metas = [
    tag_meta_b_fully_ripened,
    tag_meta_b_half_ripened,
    tag_meta_b_green,
    tag_meta_l_fully_ripened,
    tag_meta_l_half_ripened,
    tag_meta_l_green,
]

tag_meta_collection = sly.TagMetaCollection(tag_metas)
meta = sly.ProjectMeta(obj_classes=obj_class_collection, tag_metas=tag_meta_collection)

storage_dir = sly.app.get_data_dir()
work_dir_path = os.path.join(storage_dir, work_dir)
sly.io.fs.mkdir(work_dir_path)
archive_path = os.path.join(work_dir_path, arch_name)

image_name_to_id = {}
id_to_segm_anns = defaultdict(list)
id_to_tag = defaultdict(list)
name_to_size = {}
category_id_to_name = {}



from typing import Dict

import numpy as np
import supervisely as sly

from dataset_tools.image.stats.basestats import BaseStats

UNLABELED_COLOR = [0, 0, 0]


class ClassBalance(BaseStats):
    """
    Columns:
        Class
        Images
        Objects
        Avg count per image
        Avg area per image
    """

    def __init__(self, project_meta: sly.ProjectMeta, force:bool=False) -> None:
        self._meta = project_meta
        self._stats = {}
        self.force = force

        self._class_names = []
        class_colors = []
        class_indices_colors = []
        self._name_to_index = {}
        for idx, obj_class in enumerate(self._meta.obj_classes):
            self._class_names.append(obj_class.name)
            class_colors.append(obj_class.color)
            class_indices_colors.append([idx, idx, idx])
            self._name_to_index[obj_class.name] = idx

        self._stats["class_names"] = self._class_names
        self._stats["class_indices_colors"] = class_indices_colors
        self._stats["_name_to_index"] = self._name_to_index

        self._stats["sum_class_area_per_image"] = [0] * len(self._class_names)
        self._stats["objects_count"] = [0] * len(self._class_names)
        self._stats["images_count"] = [0] * len(self._class_names)

        self._stats["image_counts_filter_by_id"] = [[] for _ in self._class_names]
        # self._stats["object_counts_filter_by_id"] = [[] for _ in self._class_names]

        self._stats["avg_nonzero_area"] = [None] * len(self._class_names)
        self._stats["avg_nonzero_count"] = [None] * len(self._class_names)

    def update(self, image: sly.ImageInfo, ann: sly.Annotation):
        render_idx_rgb = np.zeros(ann.img_size + (3,), dtype=np.uint8)
        render_idx_rgb[:] = UNLABELED_COLOR

        ann.draw_class_idx_rgb(render_idx_rgb, self._stats["_name_to_index"])

        stat_area = sly.Annotation.stat_area(
            render_idx_rgb, self._stats["class_names"], self._stats["class_indices_colors"]
        )
        stat_count = ann.stat_class_count(self._stats["class_names"])

        for idx, class_name in enumerate(self._stats["class_names"]):
            cur_area = stat_area[class_name] if not np.isnan(stat_area[class_name]) else 0
            cur_count = stat_count[class_name] if not np.isnan(stat_count[class_name]) else 0

            self._stats["sum_class_area_per_image"][idx] += cur_area
            self._stats["objects_count"][idx] += cur_count
            self._stats["images_count"][idx] += 1 if stat_count[class_name] > 0 else 0

            if self._stats["images_count"][idx] > 0:
                self._stats["avg_nonzero_area"][idx] = (
                    self._stats["sum_class_area_per_image"][idx] / self._stats["images_count"][idx]
                )
                self._stats["avg_nonzero_count"][idx] = (
                    self._stats["objects_count"][idx] / self._stats["images_count"][idx]
                )

            if stat_count[class_name] > 0:
                self._stats["image_counts_filter_by_id"][idx].append(image.id)

            # TODO: implement later
            # if stat_count[class_name] > 0:
            # obj_ids = [obj[0] for obj in ann_objects if obj[1] == class_name]
            # self._stats["object_counts_filter_by_id"][idx].extend(obj_ids)

    def to_json(self) -> Dict:
        columns = [
            "Class",
            "Images",
            "Objects",
            "Count on image",
            "Area on image",
        ]
        rows = []
        for name, idx in self._name_to_index.items():
            rows.append(
                [
                    name,
                    self._stats["images_count"][idx],
                    self._stats["objects_count"][idx],
                    round(self._stats["avg_nonzero_count"][idx] or 0, 2),
                    round(self._stats["avg_nonzero_area"][idx] or 0, 2),
                ]
            )
        notnonecount = [item for item in self._stats["avg_nonzero_count"] if item is not None]
        notnonearea = [item for item in self._stats["avg_nonzero_area"] if item is not None]
        colomns_options = [None] * len(columns)
        colomns_options[0] = {"type": "class"}
        colomns_options[1] = {
            "maxValue": max(self._stats["images_count"]),
            "tooltip": "Number of images with at least one object of corresponding class",
        }
        colomns_options[2] = {
            "maxValue": max(self._stats["objects_count"]),
            "tooltip": "Number of objects of corresponding class in the project",
        }
        colomns_options[3] = {
            "maxValue": round(max(notnonecount), 2),
            "subtitle": "average",
            "tooltip": "Average number of objects of corresponding class on the image. Images without such objects are not taking into account",
        }
        colomns_options[4] = {
            "postfix": "%",
            "maxValue": round(max(notnonearea), 2),
            "subtitle": "average",
            "tooltip": "Average image area of corresponding class. Images without such objects are not taking into account",
        }
        options = {
            "fixColumns": 1,
            "sort": {"columnIndex": 1, "order": "desc"},
            "pageSize": len(self._class_names),
        }  # asc

        res = {
            "columns": columns,
            "data": rows,
            "referencesRow": self._stats["image_counts_filter_by_id"],
            "options": options,
            "columnsOptions": colomns_options,
        }
        return res



from typing import Dict

import dataframe_image as dfi
import pandas as pd

import supervisely as sly
from supervisely._utils import camel_to_snake


class BaseStats:
    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        pass

    def to_json(self) -> Dict:
        pass

    def to_pandas(self) -> pd.DataFrame:
        json = self.to_json()
        try:
            table = pd.DataFrame(data=json["data"], columns=json["columns"])
        except KeyError:
            table = None
        return table

    def to_image(self, path: str) -> None:
        """
        Create an image visualizing the results of statistics from a Pandas DataFrame.
        """
        if self.to_pandas():
            table = self.to_pandas()[:100]  # max rows == 100
            table.dfi.export(path, max_rows=-1, max_cols=-1)

    @property
    def basename_stem(self) -> str:
        """Get name of your class for your file system"""
        return camel_to_snake(self.__class__.__name__)


class BaseVisual:
    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        pass

    def to_image(
        self,
        path: str,
        draw_style: str,
        grid_spacing: int,
        outer_grid_spacing: int,
    ) -> None:
        pass

    @property
    def basename_stem(self) -> str:
        """Get name of your class for your file system"""
        return camel_to_snake(self.__class__.__name__)



from collections import defaultdict

from typing import Dict

import supervisely as sly
from supervisely.app.widgets import HeatmapChart

from dataset_tools.image.stats.basestats import BaseStats


class ObjectsDistribution(BaseStats):
    """
    Columns:
        Class
        1 object on image (if object exists)
        2 objects on image (if objects exist)
        3 objects on image (if objects exist)
        etc.
    """

    def __init__(self, project_meta: sly.ProjectMeta, force:bool = False):
        self.force = force

        self.project_meta = project_meta
        self._counters = defaultdict(lambda: {"count": 0, "image_ids": []})
        self._class_titles = [obj_class.name for obj_class in project_meta.obj_classes]
        self._data = []

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        self._data.append((image, ann))

    def to_json(self) -> Dict:
        self._stats = defaultdict(lambda: defaultdict(lambda: {"count": 0, "image_ids": []}))
        counters = defaultdict(lambda: {"count": 0, "image_ids": []})

        for image, ann in self._data:
            image_id = image.id
            counters = defaultdict(lambda: {"count": 0, "image_ids": []})

            for label in ann.labels:
                class_title = label.obj_class.name
                counters[class_title]["count"] += 1
                counters[class_title]["image_ids"].append(image_id)

            for class_title in self._class_titles:
                count = counters[class_title]["count"]
                image_ids = counters[class_title]["image_ids"]
                self._stats[class_title][count]["image_ids"].extend(list(set(image_ids)))
                self._stats[class_title][count]["count"] += 1

        max_column = max([max(class_data.keys()) for class_data in self._stats.values()])
        columns = [i for i in range(max_column + 1)]

        series = list()
        for class_title, class_data in self._stats.items():
            row = {
                "name": class_title,
                "y": [class_data[column]["count"] for column in columns],
                "x": columns,
            }

            series.append(row)

        references = defaultdict(dict)

        for column in columns:
            for class_title, class_data in self._stats.items():
                image_ids = class_data[column]["image_ids"]
                reference = {
                    column: image_ids,
                }
                if references[class_title]:
                    references[class_title].update(reference)
                else:
                    references[class_title] = reference

        hmp = HeatmapChart(
            title="Objects on images - distribution for every class",
            xaxis_title="Number of objects on image",
            color_range="row",
            tooltip="Click to preview {y} images with {x} objects of class {series_name}",
        )
        hmp.add_series_batch(series)

        res = hmp.get_json_data()
        res["referencesCell"] = references

        return res



import os
import zipfile
import tarfile


def unpack_if_archive(path: str) -> str:
    if os.path.isdir(path):
        return path

    extraction_path = os.path.splitext(path)[0]

    if zipfile.is_zipfile(path):
        os.makedirs(extraction_path, exist_ok=True)

        with zipfile.ZipFile(path, "r") as zip_ref:
            zip_ref.extractall(extraction_path)

        return extraction_path

    if tarfile.is_tarfile(path):
        os.makedirs(extraction_path, exist_ok=True)

        with tarfile.open(path, "r") as tar_ref:
            tar_ref.extractall(extraction_path)

        return extraction_path

    return path



import os
import random
import tarfile
from typing import Literal

import gdown
from cv2 import connectedComponents
from tqdm import tqdm

import dataset_tools.convert.minneapple.sly_globals as g
import supervisely as sly


def create_ann(ann_path):
    labels = []
    import cv2

    ann_np = sly.imaging.image.read(ann_path)
    ann_np_gray = cv2.cvtColor(ann_np, cv2.COLOR_BGR2GRAY)

    ann_bool = ann_np_gray != 0

    ret, curr_mask = connectedComponents(ann_bool.astype("uint8"), connectivity=8)
    for i in range(1, ret):
        obj_mask = curr_mask == i
        bitmap = sly.Bitmap(obj_mask)
        label = sly.Label(bitmap, g.obj_class)
        labels.append(label)

    return sly.Annotation(img_size=g.img_size, labels=labels)


def extract_tar():
    if tarfile.is_tarfile(g.archive_path):
        g.logger.info("Unpacking archive {}...".format(g.arch_name))
        with tarfile.open(g.archive_path, "r") as archive:
            archive.extractall(g.work_dir_path)
        g.logger.info("Archive unpacked.")
    else:
        g.logger.warn("Archive cannot be unpacked {}".format(g.arch_name))
        g.my_app.stop()


def to_supervisely(api):
    if not os.path.exists(g.archive_path):
        gdown.download(g.apple_url, g.archive_path, quiet=False)
    extract_tar()

    apple_data_path = os.path.join(g.work_dir_path, g.folder_name)

    new_project = api.project.create(g.WORKSPACE_ID, g.project_name, change_name_if_conflict=True)
    api.project.update_meta(new_project.id, g.meta.to_json())

    for ds in g.datasets:
        new_dataset = api.dataset.create(new_project.id, ds, change_name_if_conflict=True)

        curr_img_path = os.path.join(apple_data_path, ds.lower(), g.images_folder)
        curr_ann_path = os.path.join(apple_data_path, ds.lower(), g.anns_folder)

        curr_img_cnt = g.sample_img_count[ds]
        sample_img_path = random.sample(os.listdir(curr_img_path), curr_img_cnt)

        # progress = sly.Progress('Create dataset {}'.format(ds), curr_img_cnt, sly.logger)
        with tqdm(desc="Create dataset {}".format(ds), total=curr_img_cnt) as pbar:
            for img_batch in sly.batched(sample_img_path, batch_size=g.batch_size):
                img_pathes = [os.path.join(curr_img_path, name) for name in img_batch]
                img_infos = api.image.upload_paths(new_dataset.id, img_batch, img_pathes)
                img_ids = [im_info.id for im_info in img_infos]

                if ds == g.train_ds:
                    ann_pathes = [os.path.join(curr_ann_path, name) for name in img_batch]
                    anns = [create_ann(ann_path) for ann_path in ann_pathes]
                    api.annotation.upload_anns(img_ids, anns)

                pbar.update(len(img_batch))

    return new_project.id


def from_supervisely(
    input_path: str, output_path: str = None, to_format: Literal["dir", "tar", "both"] = "both"
) -> str:
    pass



from typing import Dict, List
from collections import defaultdict

import supervisely as sly
from supervisely.app.widgets import TreemapChart

from dataset_tools.image.stats.basestats import BaseStats


class ObjectSizes(BaseStats):
    """
    Columns:
        Object ID
        Class
        Dataset ID
        Image name
        Image size
        Height px
        Height %
        Width px
        Width %
        Area %
    """

    def __init__(
        self,
        project_meta: sly.ProjectMeta,
        datasets: List[sly.DatasetInfo] = None,
        force: bool = False,
    ):
        self.force = force
        self.project_meta = project_meta
        self._dataset_id_to_name = None
        if datasets is not None:
            self._dataset_id_to_name = {ds.id: ds.name for ds in datasets}
        self._stats = []
        self._object_id = 1

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        image_height, image_width = ann.img_size

        for label in ann.labels:
            if type(label.geometry) not in [sly.Bitmap, sly.Rectangle, sly.Polygon]:
                continue

            object_id = self._object_id
            self._object_id += 1

            object_data = {
                "object_id": object_id,
                "class": label.obj_class.name,
                "image_name": image.name,
            }

            if self._dataset_id_to_name:
                dataset_name = self._dataset_id_to_name[image.dataset_id]
                object_data["dataset_name"] = dataset_name

            object_data["image_size_hw"] = f"{image_height} x {image_width}"

            object_data.update(calculate_obj_sizes(label, image_height, image_width))

            object_data = list(object_data.values())

            self._stats.append((object_data, [image.id]))

    def to_json(self) -> Dict:
        options = {
            "sort": {"columnIndex": 0, "order": "asc"},
        }

        columns = [
            "Object ID",
            "Class",
            "Image name",
            "Image size",
            "Height",
            "Height",
            "Width",
            "Width",
            "Area",
        ]

        columns_options = [
            {"tooltip": "ID of the object in instance"},
            {"type": "class"},
            {"subtitle": "click row to open"},
            {"subtitle": "height x width"},
            {"postfix": "px"},
            {"postfix": "%"},
            {"postfix": "px"},
            {"postfix": "%"},
            {"postfix": "%"},
        ]

        if self._dataset_id_to_name:
            columns.insert(3, "Split")
            columns_options.insert(
                3,
                {
                    "subtitle": "folder name",
                },
            )

        data, references_row = zip(*self._stats)

        res = {
            "columns": columns,
            "columnsOptions": columns_options,
            "data": data,
            "options": options,
            "referencesRow": references_row,
        }

        return res


class ClassSizes(BaseStats):
    """
    Columns:
        Class
        Object count
        Min height px
        Min height %
        Max height px
        Max height %
        Avg height px
        Avg height %
        Min width px
        Min width %
        Max width px
        Max width %
        Avg width px
        Avg width %
        Min area px
        Min area %
        Max area px
        Max area %
        Avg area px
        Avg area %
    """

    def __init__(self, project_meta: sly.ProjectMeta, force: bool = False):
        self.force = force
        self.project_meta = project_meta
        self._class_titles = [obj_class.name for obj_class in project_meta.obj_classes]

        self._data = []

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        self._data.append(ann)

    def to_json(self) -> Dict:
        stats = []

        class_heights_px = defaultdict(list)
        class_heights_pc = defaultdict(list)
        class_widths_px = defaultdict(list)
        class_widths_pc = defaultdict(list)
        class_areas_pc = defaultdict(list)
        class_object_counts = defaultdict(int)

        for ann in self._data:
            image_height, image_width = ann.img_size
            for label in ann.labels:
                if type(label.geometry) not in [sly.Bitmap, sly.Rectangle, sly.Polygon]:
                    continue

                class_object_counts[label.obj_class.name] += 1

                obj_sizes = calculate_obj_sizes(label, image_height, image_width)

                class_heights_px[label.obj_class.name].append(obj_sizes["height_px"])
                class_heights_pc[label.obj_class.name].append(obj_sizes["height_pc"])
                class_widths_px[label.obj_class.name].append(obj_sizes["width_px"])
                class_widths_pc[label.obj_class.name].append(obj_sizes["width_pc"])
                class_areas_pc[label.obj_class.name].append(obj_sizes["area_pc"])

        for class_title in self._class_titles:
            object_count = class_object_counts[class_title]

            if object_count < 1:
                continue

            class_data = {
                "class_name": class_title,
                "object_count": object_count,
                "avg_area_pc": round(
                    sum(class_areas_pc[class_title]) / len(class_areas_pc[class_title]),
                    2,
                ),
                "max_area_pc": max(class_areas_pc[class_title]),
                "min_area_pc": min(class_areas_pc[class_title]),
                "min_height_px": min(class_heights_px[class_title]),
                "min_height_pc": min(class_heights_pc[class_title]),
                "max_height_px": max(class_heights_px[class_title]),
                "max_height_pc": max(class_heights_pc[class_title]),
                "avg_height_px": round(
                    sum(class_heights_px[class_title]) / len(class_heights_px[class_title]),
                ),
                "avg_height_pc": round(
                    sum(class_heights_pc[class_title]) / len(class_heights_pc[class_title]),
                    2,
                ),
                "min_width_px": min(class_widths_px[class_title]),
                "min_width_pc": min(class_widths_pc[class_title]),
                "max_width_px": max(class_widths_px[class_title]),
                "max_width_pc": max(class_widths_pc[class_title]),
                "avg_width_px": round(
                    sum(class_widths_px[class_title]) / len(class_widths_px[class_title]),
                ),
                "avg_width_pc": round(
                    sum(class_widths_pc[class_title]) / len(class_widths_pc[class_title]),
                    2,
                ),
            }

            class_data = list(class_data.values())

            stats.append(class_data)

        options = {
            "fixColumns": 1,
            "sort": {"columnIndex": 1, "order": "desc"},
            "pageSize": len(self._class_titles),
        }

        res = {
            "columns": [
                "Class",
                "Object count",
                "Avg area",
                "Max area",
                "Min area",
                "Min height",
                "Min height",
                "Max height",
                "Max height",
                "Avg height",
                "Avg height",
                "Min width",
                "Min width",
                "Max width",
                "Max width",
                "Avg width",
                "Avg width",
            ],
            "columnsOptions": [
                {"type": "class"},
                {"maxValue": max([class_data[1] for class_data in stats])},
                {"postfix": "%", "tooltip": "Average object area in percents of all image."},
                {"postfix": "%", "tooltip": "Maximum object area in percents of all image."},
                {"postfix": "%", "tooltip": "Minimum object area in percents of all image."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Minimum object height in percents of image height."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Maximum object height in percents of image height."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Average object height in percents of image height."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Minimum object width in percents of image width."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Maximum object width in percents of image width."},
                {"postfix": "px"},
                {"postfix": "%", "tooltip": "Average object width in percents of image width."},
            ],
            "data": stats,
            "options": options,
        }

        return res


class ClassTreemap(BaseStats):
    def __init__(self, project_meta: sly.ProjectMeta, force: bool = False):
        self.force = force
        self.project_meta = project_meta
        self._class_titles = [obj_class.name for obj_class in project_meta.obj_classes]
        self._class_rgbs = [obj_class.color for obj_class in project_meta.obj_classes]
        self._class_colors = [rgb_to_hex(rgb) for rgb in self._class_rgbs]

        self._data = []

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        self._data.append(ann)

    def to_json(self) -> Dict:
        tooltip = "Average area of class objects on image is {y}%"
        colors = self._class_colors
        names = []
        values = []

        class_areas_pc = defaultdict(list)
        class_object_counts = defaultdict(int)

        for ann in self._data:
            image_height, image_width = ann.img_size
            for label in ann.labels:
                if type(label.geometry) not in [sly.Bitmap, sly.Rectangle, sly.Polygon]:
                    continue

                class_object_counts[label.obj_class.name] += 1

                obj_sizes = calculate_obj_sizes(label, image_height, image_width)

                class_areas_pc[label.obj_class.name].append(obj_sizes["area_pc"])

        for class_title in self._class_titles:
            object_count = class_object_counts[class_title]

            if object_count < 1:
                continue

            names.append(class_title)
            values.append(
                round(
                    sum(class_areas_pc[class_title]) / len(class_areas_pc[class_title]),
                    2,
                )
            )

        tc = TreemapChart(
            title="Average area of class objects on image",
            colors=colors,
            tooltip=tooltip,
        )

        tc.set_series(names, values)

        res = tc.get_json_data()

        return res


def rgb_to_hex(rgb: List[int]) -> str:
    return "#{:02x}{:02x}{:02x}".format(*rgb)


def calculate_obj_sizes(label: sly.Label, image_height: int, image_width: int) -> Dict:
    image_area = image_height * image_width

    rect_geometry = label.geometry.to_bbox()

    height_px = rect_geometry.height
    height_pc = round(height_px * 100.0 / image_height, 2)

    width_px = rect_geometry.width
    width_pc = round(width_px * 100.0 / image_width, 2)

    area_px = label.geometry.area
    area_pc = round(area_px * 100.0 / image_area, 2)

    return {
        "height_px": height_px,
        "height_pc": height_pc,
        "width_px": width_px,
        "width_pc": width_pc,
        "area_pc": area_pc,
    }



import os
import sys
from pathlib import Path

import supervisely as sly

# my_app = sly.AppService()
# api: sly.Api = my_app.public_api

root_source_dir = str(Path(sys.argv[0]).parents[1])
sly.logger.info(f"Root source directory: {root_source_dir}")
sys.path.append(root_source_dir)

# TASK_ID = int(os.environ["TASK_ID"])
TEAM_ID = sly.env.team_id()
WORKSPACE_ID = sly.env.workspace_id()

logger = sly.logger

# train_ds = os.environ["modal.state.train"]
# test_ds = os.environ["modal.state.test"]

datasets = ["Train", "Test"]

# for ds in [train_ds, test_ds]:
#     if len(ds) != 2:
#         datasets.append(ds[1:-1].replace("'", ""))

# if len(datasets) == 0:
#     logger.warn("You have not selected a dataset to import")
#     my_app.stop()

train_percent = 100
test_percent = 100

sample_img_count = {"Train": round(6.7 * train_percent), "Test": round(3.31 * test_percent)}

project_name = "MinneApple"
work_dir = "apple_data"
apple_url = "https://conservancy.umn.edu/bitstream/handle/11299/206575/detection.tar.gz?sequence=2&isAllowed=y"

arch_name = "detection.tar.gz"
folder_name = "detection"
images_folder = "images"
anns_folder = "masks"
img_size = (1280, 720)
batch_size = 30
class_name = "apple"
train_ds = "Train"

obj_class = sly.ObjClass(class_name, sly.Bitmap)
obj_class_collection = sly.ObjClassCollection([obj_class])

meta = sly.ProjectMeta(obj_classes=obj_class_collection)

storage_dir = sly.app.get_data_dir()
work_dir_path = os.path.join(storage_dir, work_dir)
sly.io.fs.mkdir(work_dir_path)
archive_path = os.path.join(work_dir_path, arch_name)



from collections import defaultdict
from typing import Dict

import numpy as np
import supervisely as sly

from dataset_tools.image.stats.basestats import BaseStats

# from supervisely.app.widgets import ConfusionMatrix


class ClassCooccurrence(BaseStats):
    """
    Columns:
        Class
        class 1
        class 2
        etc.
    """

    def __init__(self, project_meta: sly.ProjectMeta, force:bool = False) -> None:
        self._meta = project_meta
        self._stats = {}
        self.force = force

        self._name_to_index = {}

        for idx, obj_class in enumerate(self._meta.obj_classes):
            self._name_to_index[obj_class.name] = idx

        self._class_names = [cls.name for cls in project_meta.obj_classes]
        self._references = defaultdict(lambda: defaultdict(list))

        num_classes = len(self._class_names)
        self.co_occurrence_matrix = np.zeros((num_classes, num_classes), dtype=int)

    def update(self, image: sly.ImageInfo, ann: sly.Annotation) -> None:
        classes = set()
        for label in ann.labels:
            classes.add(label.obj_class.name)

        for class_ in classes:
            idx = self._name_to_index[class_]
            self.co_occurrence_matrix[idx][idx] += 1
            self._references[idx][idx].append(image.id)

        classes = list(classes)
        for i in range(len(classes)):
            for j in range(i + 1, len(classes)):
                class_i = classes[i]
                class_j = classes[j]
                idx_i = self._name_to_index[class_i]
                idx_j = self._name_to_index[class_j]
                self.co_occurrence_matrix[idx_i][idx_j] += 1
                self.co_occurrence_matrix[idx_j][idx_i] += 1

                self._references[idx_i][idx_j].append(image.id)
                self._references[idx_j][idx_i].append(image.id)

    def to_json(self) -> Dict:
        options = {
            "fixColumns": 1,  # not used in Web
            "cellTooltip": "Click to preview. {currentCell} images have objects of both classes {firstCell} and {currentColumn} at the same time",
        }
        colomns_options = [None] * (len(self._class_names) + 1)
        colomns_options[0] = {"type": "class"}  # not used in Web

        for idx in range(1, len(colomns_options)):
            colomns_options[idx] = {"maxValue": int(np.max(self.co_occurrence_matrix[:, idx - 1]))}

        data = [
            [value] + sublist
            for value, sublist in zip(self._class_names, self.co_occurrence_matrix.tolist())
        ]

        res = {
            "columns": ["Class"] + self._class_names,
            "data": data,
            "referencesCell": self._references,  # row - col
            "options": options,
            "colomnsOptions": colomns_options,
        }
        return res

    # def get_widget(self) -> ConfusionMatrix:
    #     df = pd.DataFrame(data=self.co_occurrence_matrix.tolist(), columns=self._class_names)
    #     confusion_matrix = ConfusionMatrix()
    #     confusion_matrix.read_pandas(df)
    #     return confusion_matrix



import os
import random
from typing import Union, List

from tqdm import tqdm

import supervisely as sly


def sample_images(
    api: sly.Api,
    project: Union[int, str],
    datasets: List[Union[sly.DatasetInfo, sly.Project.DatasetDict]],
    sample_rate: float,
):
    total = 0
    samples = []
    for dataset in datasets:
        k = int(
            max(
                1,
                sample_rate
                * (
                    dataset.items_count
                    if isinstance(project, int)
                    else len(os.listdir(dataset.ann_dir))
                ),
            )
        )

        ds_images = (
            api.image.get_list(dataset.id)
            if isinstance(project, int)
            else [
                dataset.get_image_info(sly.fs.get_file_name(img))
                for img in os.listdir(dataset.ann_dir)
            ]
        )

        s = random.sample(ds_images, k)
        samples.append((dataset, s))
        total += k
    return samples, total


def count_stats(
    project: Union[int, str], stats: list, sample_rate: float = 1, api: sly.Api = None
) -> None:
    """
    Count dtools statistics instances passed as a list.

    :param project: Supervisely project ID or a local project path.
    :type project: Union[int, str]
    :param stats: list of instances of statistics
    :type stats: list
    :param sample_rate: Modify size of a statistics sample.
    :type sample_rate: float, optional
    :param api: Supervisely API
    :type api: sly.Api, optional

    :Usage example:

     .. code-block:: python

        import os
        from dotenv import load_dotenv

        import supervisely as sly

        # Load secrets and create API object from .env file (recommended)
        # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
        if sly.is_development():
           load_dotenv(os.path.expanduser("~/supervisely.env"))
        api = sly.Api.from_env()

        # Pass values into the API constructor (optional, not recommended)
        # api = sly.Api(server_address="https://app.supervise.ly", token="4r47N...xaTatb")

        api = sly.Api.from_env()

        project_id = sly.env.project_id()
        project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project_id))
        datasets = api.dataset.get_list(project_id)

        stats = [
            dtools.ClassesPerImage(project_meta, datasets),
            dtools.ClassBalance(project_meta),
            dtools.ClassCooccurrence(project_meta),
        ]
        dtools.count_stats(
            project_id,
            stats=stats,
            sample_rate=0.01,
        )
        print("Saving stats...")
        for stat in stats:
            with open(f"./stats/{stat.basename_stem}.json", "w") as f:
                json.dump(stat.to_json(), f)
            stat.to_image(f"./stats/{stat.basename_stem}.png")
        print("Done")
    """
    if len(stats) == 0:
        print("Passed 'stats' parameter is empty. Enable 'force' flag to overwrite statistics output file. Skipping statistics counting...")
        return
    if sample_rate <= 0 or sample_rate > 1:
        raise ValueError("Sample rate has to be in range (0, 1]")
    if api is None:
        api = sly.Api.from_env()

    if isinstance(project, int):
        project_meta = sly.ProjectMeta.from_json(api.project.get_meta(project))
        datasets = api.dataset.get_list(project)
    elif isinstance(project, str):
        project_fs = sly.Project(project, sly.OpenMode.READ)
        project_meta = project_fs.meta
        datasets = project_fs.datasets
    else:
        raise ValueError("Project should be either an integer project ID or a string project path.")

    samples, total = sample_images(api, project, datasets, sample_rate)
    desc = "Calculating stats" + (f" [sample={sample_rate}]" if sample_rate != 1 else "")
    with tqdm(desc=desc, total=total) as pbar:
        for dataset, images in samples:
            for batch in sly.batched(images):
                image_ids = [image.id for image in batch]
                image_names = [image.name for image in batch]

                if isinstance(project, int):
                    janns = api.annotation.download_json_batch(dataset.id, [id for id in image_ids])
                    anns = [sly.Annotation.from_json(ann_json, project_meta) for ann_json in janns]
                else:
                    anns = [dataset.get_ann(name, project_meta) for name in image_names]

                for img, ann in zip(batch, anns):
                    for stat in stats:
                        stat.update(img, ann)
                    pbar.update(1)



import subprocess
from subprocess import PIPE


def from_mp4_to_webm(src_path, dst_path, compression_level: int = 35):
    session = subprocess.Popen(
        [
            "ffmpeg",
            "-y",
            "-i",
            f"{src_path}",
            "-vf",
            "scale='min(1600,iw)':-2",
            "-vcodec",
            "libvpx-vp9",
            "-crf",
            f"{compression_level}",
            "-b:v",
            "0",
            "-threads",
            "4",
            "-cpu-used",
            "4",
            "-pix_fmt",
            "yuv420p",
            "-an",
            f"{dst_path}",
        ],
        stdout=PIPE,
        stderr=PIPE,
    )
    stdout, stderr = session.communicate()
    return dst_path


def compress_mp4(src_path, dst_path, compression_level: int = 35):
    session = subprocess.Popen(
        [
            "ffmpeg",
            "-y",
            "-i",
            f"{src_path}",
            "-vf",
            "scale='min(1600,iw)':-2",
            "-c:v",
            "libx264",
            "-crf",
            f"{compression_level}",
            "-b:v",
            "0",
            "-pix_fmt",
            "yuv420p",
            "-an",
            f"{dst_path}",
        ],
        stdout=PIPE,
        stderr=PIPE,
    )
    stdout, stderr = session.communicate()
    return dst_path


def compress_png(src_path, dst_path, resolution: int = 720):
    session = subprocess.Popen(
        [
            "ffmpeg",
            "-y",
            "-ss",
            "0",
            "-i",
            f"{src_path}",
            "-filter:v",
            f"scale=-2:{resolution}",
            f"{dst_path}",
        ],
        stdout=PIPE,
        stderr=PIPE,
    )
    stdout, stderr = session.communicate()
    return dst_path



import os
import json

from collections import OrderedDict
from shutil import copyfile

from dataset_tools.convert import unpack_if_archive

import numpy as np
import lxml.etree as ET
from PIL import Image
import supervisely as sly
from supervisely.io.fs import get_file_name
from supervisely.imaging.color import generate_rgb


default_classes_colors = {
    "neutral": (224, 224, 192),
    "aeroplane": (128, 0, 0),
    "bicycle": (0, 128, 0),
    "bird": (128, 128, 0),
    "boat": (0, 0, 128),
    "bottle": (128, 0, 128),
    "bus": (0, 128, 128),
    "car": (128, 128, 128),
    "cat": (64, 0, 0),
    "chair": (192, 0, 0),
    "cow": (64, 128, 0),
    "diningtable": (192, 128, 0),
    "dog": (64, 0, 128),
    "horse": (192, 0, 128),
    "motorbike": (64, 128, 128),
    "person": (192, 128, 128),
    "pottedplant": (0, 64, 0),
    "sheep": (128, 64, 0),
    "sofa": (0, 192, 0),
    "train": (128, 192, 0),
    "tvmonitor": (0, 64, 128),
}

MASKS_EXTENSION = ".png"


def to_supervisely(input_path: str, output_path: str = None):
    input_path = unpack_if_archive(input_path)
    # Specific directory that must exist in input path.
    DIR_NAME = "VOCdevkit"

    dataset_dir = os.path.join(input_path, DIR_NAME)

    if not os.path.exists(dataset_dir):
        raise FileNotFoundError(
            f"Input directory must contain {DIR_NAME} directory, but it is not found in {input_path}."
        )

    if not output_path:
        output_path = os.path.join(os.path.dirname(input_path), "PASCAL_TO_SLY")

    # Rename VOC to VOC2012 if necessary.
    if not os.path.isdir(os.path.join(dataset_dir, "VOC2012")):
        try:
            os.rename(os.path.join(dataset_dir, "VOC"), os.path.join(dataset_dir, "VOC2012"))
        except FileNotFoundError:
            raise FileNotFoundError(
                f"Input directory must contain {DIR_NAME}/VOC or {DIR_NAME}/VOC2012 directory, "
                f"but it is not found in {input_path}."
            )

    lists_dir = os.path.join(dataset_dir, "VOC2012", "ImageSets", "Segmentation")
    imgs_dir = os.path.join(dataset_dir, "VOC2012", "JPEGImages")
    segm_dir = os.path.join(dataset_dir, "VOC2012", "SegmentationClass")
    inst_dir = os.path.join(dataset_dir, "VOC2012", "SegmentationObject")
    colors_file = os.path.join(dataset_dir, "VOC2012", "colors.txt")
    with_instances = os.path.isdir(inst_dir)

    obj_classes = sly.ObjClassCollection()

    src_datasets = {}
    if not os.path.isdir(lists_dir):
        raise RuntimeError(f"There is no directory {lists_dir}, but it is necessary")

    for filename in os.listdir(lists_dir):
        if filename.endswith(".txt"):
            ds_name = os.path.splitext(filename)[0]
            file_path = os.path.join(lists_dir, filename)
            sample_names = list(filter(None, map(str.strip, open(file_path, "r").readlines())))
            src_datasets[ds_name] = sample_names

    if os.path.isfile(colors_file):
        in_lines = filter(None, map(str.strip, open(colors_file, "r").readlines()))
        in_splitted = (x.split() for x in in_lines)
        cls2col = {x[0]: (int(x[1]), int(x[2]), int(x[3])) for x in in_splitted}
    else:
        cls2col = default_classes_colors

    obj_classes_list = [
        sly.ObjClass(name=class_name, geometry_type=sly.Bitmap, color=color)
        for class_name, color in cls2col.items()
    ]

    obj_classes = obj_classes.add_items(obj_classes_list)
    color2class_name = {v: k for k, v in cls2col.items()}

    out_project = sly.Project(output_path, sly.OpenMode.CREATE)
    images_filenames = {}

    for image_path in sly.fs.list_files(imgs_dir):
        image_name_noext = sly.fs.get_file_name(image_path)
        if image_name_noext in images_filenames:
            raise RuntimeError(
                "Multiple image with the same base name {!r} exist.".format(image_name_noext)
            )
        images_filenames[image_name_noext] = image_path

    for ds_name, sample_names in src_datasets.items():
        if len(sample_names) == 0:
            continue
        ds = out_project.create_dataset(ds_name)
        percent_counter = 0

        for sample_name in sample_names:
            percent_counter += 1
            try:
                src_img_path = images_filenames[get_file_name(sample_name)]
            except Exception:
                src_img_path = images_filenames[sample_name]
            src_img_filename = os.path.basename(src_img_path)
            segm_path = os.path.join(segm_dir, sample_name + MASKS_EXTENSION)

            inst_path = None
            if with_instances:
                inst_path = os.path.join(inst_dir, sample_name + MASKS_EXTENSION)

            if all((x is None) or os.path.isfile(x) for x in [src_img_path, segm_path, inst_path]):
                try:
                    ann = get_ann(src_img_path, segm_path, inst_path, color2class_name)
                    ds.add_item_file(src_img_filename, src_img_path, ann=ann)
                except Exception as e:
                    exc_str = str(e)
                    sly.logger.warn(
                        f"Input sample skipped due to error: {exc_str}",
                        exc_info=True,
                        extra={
                            "exc_str": exc_str,
                            "dataset_name": ds_name,
                            "image": src_img_path,
                        },
                    )

            else:
                ds.add_item_file(src_img_filename, src_img_path, ann=None)

    out_meta = sly.ProjectMeta(obj_classes=obj_classes)
    out_project.set_meta(out_meta)

    return output_path


def get_ann(img_path, segm_path, inst_path, color2class_name):
    segmentation_img = sly.image.read(segm_path)

    if inst_path is not None:
        instance_img = sly.image.read(inst_path)
        colored_img = instance_img
        instance_img16 = instance_img.astype(np.uint16)
        col2coord = get_col2coord(instance_img16)
        curr_col2cls = (
            (col, color2class_name.get(tuple(segmentation_img[coord])))
            for col, coord in col2coord.items()
        )
        curr_col2cls = {
            k: v for k, v in curr_col2cls if v is not None
        }  # _instance_ color -> class name
    else:
        colored_img = segmentation_img
        segmentation_img = segmentation_img.astype(np.uint16)
        colors = list(get_col2coord(segmentation_img).keys())
        curr_col2cls = {curr_col: color2class_name[curr_col] for curr_col in colors}

    ann = sly.Annotation.from_img_path(img_path)

    for color, class_name in curr_col2cls.items():
        mask = np.all(colored_img == color, axis=2)  # exact match (3-channel img & rgb color)
        bitmap = sly.Bitmap(data=mask)
        obj_class = sly.ObjClass(name=class_name, geometry_type=sly.Bitmap)

        ann = ann.add_label(sly.Label(bitmap, obj_class))
        #  clear used pixels in mask to check missing colors, see below
        colored_img[mask] = (0, 0, 0)

    if np.sum(colored_img) > 0:
        sly.logger.warn("Not all objects or classes are captured from source segmentation.")

    return ann


# returns mapping: (r, g, b) color -> some (row, col) for each unique color except black
def get_col2coord(img):
    img = img.astype(np.int32)
    h, w = img.shape[:2]
    colhash = img[:, :, 0] * 256 * 256 + img[:, :, 1] * 256 + img[:, :, 2]
    unq, unq_inv, unq_cnt = np.unique(colhash, return_inverse=True, return_counts=True)
    indxs = np.split(np.argsort(unq_inv), np.cumsum(unq_cnt[:-1]))
    col2indx = {unq[i]: indxs[i][0] for i in range(len(unq))}
    return {
        (col // (256**2), (col // 256) % 256, col % 256): (indx // w, indx % w)
        for col, indx in col2indx.items()
        if col != 0
    }


###################
PASCAL_CONTOUR_THICKNESS = 3
TRAIN_VAL_SPLIT_COEF = 0.8
RESULT_SUBDIR_NAME = "VOCdevkit/VOC"
VALID_IMG_EXT = set([".jpe", ".jpeg", ".jpg"])

TRAIN_TAG_NAME = "train"
VAL_TAG_NAME = "val"
SPLIT_TAGS = set([TRAIN_TAG_NAME, VAL_TAG_NAME])

SUPPORTED_GEOMETRY_TYPES = set([sly.Bitmap, sly.Polygon])

images_dir_name = "JPEGImages"
ann_dir_name = "Annotations"
ann_class_dir_name = "SegmentationClass"
ann_obj_dir_name = "SegmentationObject"

trainval_sets_dir_name = "ImageSets"
trainval_sets_main_name = "Main"
trainval_sets_segm_name = "Segmentation"

pascal_contour_color = [224, 224, 192]
pascal_ann_ext = ".png"


def from_supervisely(input_path: str, output_path: str = None):
    input_path = unpack_if_archive(input_path)
    project_info = sly.Project(input_path, sly.OpenMode.READ)
    meta = project_info.meta

    if not output_path:
        output_path = os.path.join(os.path.dirname(input_path), "SLY_TO_PASCAL")

    result_dir = output_path
    result_subdir = os.path.join(result_dir, RESULT_SUBDIR_NAME)

    result_ann_dir = os.path.join(result_subdir, ann_dir_name)
    result_images_dir = os.path.join(result_subdir, images_dir_name)
    result_class_dir_name = os.path.join(result_subdir, ann_class_dir_name)
    result_obj_dir = os.path.join(result_subdir, ann_obj_dir_name)
    result_imgsets_dir = os.path.join(result_subdir, trainval_sets_dir_name)

    sly.fs.mkdir(result_ann_dir)
    sly.fs.mkdir(result_imgsets_dir)
    sly.fs.mkdir(result_images_dir)
    sly.fs.mkdir(result_class_dir_name)
    sly.fs.mkdir(result_obj_dir)

    images_stats = []
    classes_colors = {}

    datasets = project_info.datasets
    dataset_names = ["trainval", "val", "train"]

    for dataset in datasets:
        if dataset.name in dataset_names:
            is_trainval = 1
        else:
            is_trainval = 0

        image_names = [f for f in os.listdir(dataset.img_dir)]
        src_image_paths = [os.path.join(dataset.img_dir, f) for f in image_names]
        anns_json = [
            json.load(open(os.path.join(dataset.ann_dir, f))) for f in os.listdir(dataset.ann_dir)
        ]
        dst_image_paths = [os.path.join(result_images_dir, f) for f in image_names]

        for src, dst in zip(src_image_paths, dst_image_paths):
            copyfile(src, dst)

        for image_name, image_path, ann_json in zip(image_names, dst_image_paths, anns_json):
            img_title, img_ext = os.path.splitext(image_name)
            cur_img_filename = image_name

            if is_trainval == 1:
                cur_img_stats = {"classes": set(), "dataset": dataset.name, "name": img_title}
                images_stats.append(cur_img_stats)
            else:
                cur_img_stats = {"classes": set(), "dataset": None, "name": img_title}
                images_stats.append(cur_img_stats)

            if img_ext not in VALID_IMG_EXT:
                orig_image_path = os.path.join(result_images_dir, cur_img_filename)

                jpg_image = img_title + ".jpg"
                jpg_image_path = os.path.join(result_images_dir, jpg_image)

                im = sly.image.read(orig_image_path)
                sly.image.write(jpg_image_path, im)
                sly.fs.silent_remove(orig_image_path)

            ann = sly.Annotation.from_json(ann_json, meta)
            tag = find_first_tag(ann.img_tags, SPLIT_TAGS)
            if tag is not None:
                cur_img_stats["dataset"] = tag.meta.name

            valid_labels = []
            for label in ann.labels:
                if type(label.geometry) in SUPPORTED_GEOMETRY_TYPES:
                    valid_labels.append(label)

            ann = ann.clone(labels=valid_labels)
            ann_to_xml(project_info, image_path, cur_img_filename, result_ann_dir, ann)
            for label in ann.labels:
                cur_img_stats["classes"].add(label.obj_class.name)
                classes_colors[label.obj_class.name] = tuple(label.obj_class.color)

            fake_contour_th = 0
            if PASCAL_CONTOUR_THICKNESS != 0:
                fake_contour_th = 2 * PASCAL_CONTOUR_THICKNESS + 1

            from_ann_to_instance_mask(
                ann,
                os.path.join(result_class_dir_name, img_title + pascal_ann_ext),
                fake_contour_th,
            )
            from_ann_to_class_mask(
                ann, os.path.join(result_obj_dir, img_title + pascal_ann_ext), fake_contour_th
            )

    classes_colors = OrderedDict((sorted(classes_colors.items(), key=lambda t: t[0])))

    with open(os.path.join(result_subdir, "colors.txt"), "w") as cc:
        if PASCAL_CONTOUR_THICKNESS != 0:
            cc.write(
                f"neutral {pascal_contour_color[0]} {pascal_contour_color[1]} {pascal_contour_color[2]}\n"
            )

        for k in classes_colors.keys():
            if k == "neutral":
                continue

            cc.write(f"{k} {classes_colors[k][0]} {classes_colors[k][1]} {classes_colors[k][2]}\n")

    imgs_to_split = [i for i in images_stats if i["dataset"] is None]
    train_len = int(len(imgs_to_split) * TRAIN_VAL_SPLIT_COEF)

    for img_stat in imgs_to_split[:train_len]:
        img_stat["dataset"] = TRAIN_TAG_NAME
    for img_stat in imgs_to_split[train_len:]:
        img_stat["dataset"] = VAL_TAG_NAME

    write_segm_set(is_trainval, images_stats, result_imgsets_dir)
    write_main_set(is_trainval, images_stats, meta, result_imgsets_dir)

    return output_path


def find_first_tag(img_tags, split_tags):
    for tag in split_tags:
        if img_tags.has_key(tag):
            return img_tags.get(tag)
    return None


def from_ann_to_instance_mask(ann, mask_outpath, contour_thickness):
    mask = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
    for label in ann.labels:
        if label.obj_class.name == "neutral":
            label.geometry.draw(mask, pascal_contour_color)
            continue

        label.geometry.draw_contour(mask, pascal_contour_color, contour_thickness)
        label.geometry.draw(mask, label.obj_class.color)

    im = Image.fromarray(mask)
    im = im.convert("P", palette=Image.ADAPTIVE)
    im.save(mask_outpath)


def from_ann_to_class_mask(ann, mask_outpath, contour_thickness):
    exist_colors = [[0, 0, 0], pascal_contour_color]
    mask = np.zeros((ann.img_size[0], ann.img_size[1], 3), dtype=np.uint8)
    for label in ann.labels:
        if label.obj_class.name == "neutral":
            label.geometry.draw(mask, pascal_contour_color)
            continue

        new_color = generate_rgb(exist_colors)
        exist_colors.append(new_color)
        label.geometry.draw_contour(mask, pascal_contour_color, contour_thickness)
        label.geometry.draw(mask, new_color)

    im = Image.fromarray(mask)
    im = im.convert("P", palette=Image.ADAPTIVE)
    im.save(mask_outpath)


def write_main_set(is_trainval, images_stats, meta_json, result_imgsets_dir):
    result_imgsets_main_subdir = os.path.join(result_imgsets_dir, trainval_sets_main_name)
    result_imgsets_segm_subdir = os.path.join(result_imgsets_dir, trainval_sets_segm_name)
    sly.fs.mkdir(result_imgsets_main_subdir)

    res_files = ["trainval.txt", "train.txt", "val.txt"]
    for file in os.listdir(result_imgsets_segm_subdir):
        if file in res_files:
            copyfile(
                os.path.join(result_imgsets_segm_subdir, file),
                os.path.join(result_imgsets_main_subdir, file),
            )

    train_imgs = [i for i in images_stats if i["dataset"] == TRAIN_TAG_NAME]
    val_imgs = [i for i in images_stats if i["dataset"] == VAL_TAG_NAME]

    write_objs = [
        {"suffix": "trainval", "imgs": images_stats},
        {"suffix": "train", "imgs": train_imgs},
        {"suffix": "val", "imgs": val_imgs},
    ]

    if is_trainval == 1:
        trainval_imgs = [i for i in images_stats if i["dataset"] == TRAIN_TAG_NAME + VAL_TAG_NAME]
        write_objs[0] = {"suffix": "trainval", "imgs": trainval_imgs}

    for obj_cls in meta_json.obj_classes:
        if obj_cls.geometry_type not in SUPPORTED_GEOMETRY_TYPES:
            continue
        if obj_cls.name == "neutral":
            continue
        for o in write_objs:
            with open(
                os.path.join(result_imgsets_main_subdir, f'{obj_cls.name}_{o["suffix"]}.txt'), "w"
            ) as f:
                for img_stats in o["imgs"]:
                    v = "1" if obj_cls.name in img_stats["classes"] else "-1"
                    f.write(f'{img_stats["name"]} {v}\n')


def write_segm_set(is_trainval, images_stats, result_imgsets_dir):
    result_imgsets_segm_subdir = os.path.join(result_imgsets_dir, trainval_sets_segm_name)
    sly.fs.mkdir(result_imgsets_segm_subdir)

    with open(os.path.join(result_imgsets_segm_subdir, "trainval.txt"), "w") as f:
        if is_trainval == 1:
            f.writelines(
                i["name"] + "\n"
                for i in images_stats
                if i["dataset"] == TRAIN_TAG_NAME + VAL_TAG_NAME
            )
        else:
            f.writelines(i["name"] + "\n" for i in images_stats)
    with open(os.path.join(result_imgsets_segm_subdir, "train.txt"), "w") as f:
        f.writelines(i["name"] + "\n" for i in images_stats if i["dataset"] == TRAIN_TAG_NAME)
    with open(os.path.join(result_imgsets_segm_subdir, "val.txt"), "w") as f:
        f.writelines(i["name"] + "\n" for i in images_stats if i["dataset"] == VAL_TAG_NAME)


def ann_to_xml(project_info, image_path, img_filename, result_ann_dir, ann):
    xml_root = ET.Element("annotation")

    ET.SubElement(xml_root, "folder").text = "VOC_" + project_info.name
    ET.SubElement(xml_root, "filename").text = img_filename

    xml_root_source = ET.SubElement(xml_root, "source")
    ET.SubElement(xml_root_source, "database").text = "Supervisely Project ID:" + str(
        project_info.name
    )
    ET.SubElement(xml_root_source, "annotation").text = "PASCAL VOC"
    # ET.SubElement(xml_root_source, "image").text = "Supervisely Image ID:" + str(image_info.id)
    ET.SubElement(xml_root_source, "image").text = "Supervisely Image ID:" + img_filename

    image = Image.open(image_path)
    width, height = image.size

    xml_root_size = ET.SubElement(xml_root, "size")
    # ET.SubElement(xml_root_size, "width").text = str(image_info.width)
    # ET.SubElement(xml_root_size, "height").text = str(image_info.height)
    ET.SubElement(xml_root_size, "width").text = str(width)
    ET.SubElement(xml_root_size, "height").text = str(height)
    ET.SubElement(xml_root_size, "depth").text = "3"

    ET.SubElement(xml_root, "segmented").text = "1" if len(ann.labels) > 0 else "0"

    for label in ann.labels:
        if label.obj_class.name == "neutral":
            continue

        bitmap_to_bbox = label.geometry.to_bbox()

        xml_ann_obj = ET.SubElement(xml_root, "object")
        ET.SubElement(xml_ann_obj, "name").text = label.obj_class.name
        ET.SubElement(xml_ann_obj, "pose").text = "Unspecified"
        ET.SubElement(xml_ann_obj, "truncated").text = "0"
        ET.SubElement(xml_ann_obj, "difficult").text = "0"

        xml_ann_obj_bndbox = ET.SubElement(xml_ann_obj, "bndbox")
        ET.SubElement(xml_ann_obj_bndbox, "xmin").text = str(bitmap_to_bbox.left)
        ET.SubElement(xml_ann_obj_bndbox, "ymin").text = str(bitmap_to_bbox.top)
        ET.SubElement(xml_ann_obj_bndbox, "xmax").text = str(bitmap_to_bbox.right)
        ET.SubElement(xml_ann_obj_bndbox, "ymax").text = str(bitmap_to_bbox.bottom)

    tree = ET.ElementTree(xml_root)

    img_name = os.path.join(result_ann_dir, os.path.splitext(img_filename)[0] + ".xml")
    ann_path = os.path.join(result_ann_dir, img_name)
    ET.indent(tree, space="    ")
    tree.write(ann_path, pretty_print=True)



import os
import random
from typing import Union

import cv2
import numpy as np
from tqdm import tqdm

import supervisely as sly
from dataset_tools.image.renders.convert import from_mp4_to_webm, compress_mp4, compress_png


class VerticalGrid:
    def __init__(
        self,
        project: Union[str, int],
        project_meta: sly.ProjectMeta,
        api: sly.Api = None,
        rows: int = 6,
        cols: int = 3,
        footer_path: str = "dninja_footer.png",
        force: bool = False,
    ):
        self.force = force
        self.project_meta = project_meta

        self._img_width = 1920
        self._rows = rows
        self._cols = cols
        self._g_spacing = 15
        self._column_height = 0
        self._footer_path = footer_path

        self.np_images = []  # for grid
        self.np_anns = []  # for gif
        self.np_frames = []  # for gif
        self._img_array = None
        self._column_width = int(
            (self._img_width - self._g_spacing * (self._cols + 1)) / self._cols
        )

        self._local = False if isinstance(project, int) else True
        self._api = api if api is not None else sly.Api.from_env()

    @property
    def basename_stem(self) -> None:
        return sly.utils.camel_to_snake(self.__class__.__name__)

    def update(self, data: tuple):
        cnt = self._cols * self._rows
        join_data = [(ds, img, ann) for ds, list1, list2 in data for img, ann in zip(list1, list2)]

        random.shuffle(join_data)
        with tqdm(desc="Downloading images", total=cnt) as p:
            for ds, img_info, ann in join_data[:cnt]:
                ann: sly.Annotation
                img = (
                    sly.image.read(ds.get_img_path(img_info.name))
                    if self._local
                    else self._api.image.download_np(img_info.id)
                )
                tmp = np.dstack((img, np.ones((*img.shape[:2], 1), dtype=np.uint8) * 255))
                ann_mask = np.ones((*img.shape[:2], 4), dtype=np.uint8) * 255
                ann.draw_pretty(ann_mask[:, :, :3], thickness=0, opacity=0.7)
                self.np_frames.append(self._resize_image(tmp, self._column_width))  # for gif
                self.np_anns.append(self._resize_image(ann_mask, self._column_width))  # for gif

                ann.draw_pretty(img, thickness=0, opacity=0.7)
                img = self._resize_image(img, self._column_width)
                self.np_images.append(img)

                p.update(1)

    def to_image(self, path: str = None):
        path_part, ext = os.path.splitext(path)
        tmp_path = f"{path_part}-o{ext}"
        if path is None:
            storage_dir = sly.app.get_data_dir()
            path = os.path.join(storage_dir, "vertical_grid.png")
        self._img_array = self._merge_canvas_with_images(self.np_images)
        self._add_footer_with_logo(self._img_array)

        sly.image.write(tmp_path, self._img_array)
        compress_png(tmp_path, path, 1080)
        sly.fs.silent_remove(tmp_path)
        sly.logger.info(f"Result grid saved to: {path}")

    def animate(self, path: str = None):
        bg = self._merge_canvas_with_images(self.np_frames, 4)
        ann = self._merge_canvas_with_images(self.np_anns, 4)
        ann[:, :, 3] = np.where(np.all(ann == 255, axis=-1), 0, 255).astype(np.uint8)

        duration = 1.1
        fps = 15
        num_frames = int(duration * fps)
        frames = []
        for i in list(range(1, num_frames + 1)) + list(range(num_frames, 0, -1)):
            alpha = i / num_frames
            frame = self._overlay_images(bg, ann, alpha)
            self._add_footer_with_logo(frame)
            frame = self._resize_image(frame, frame.shape[1] // 2)
            frames.append(frame)
            if i == num_frames:
                frames.extend([frame] * (fps // 2))

        tmp_video_path = f"{os.path.splitext(path)[0]}-o.mp4"
        video_path = f"{os.path.splitext(path)[0]}.mp4"
        self._save_video(tmp_video_path, frames)
        from_mp4_to_webm(tmp_video_path, path)
        compress_mp4(tmp_video_path, video_path)
        sly.fs.silent_remove(tmp_video_path)

        sly.logger.info(f"Animation saved to: {path}, {video_path}")

    def _merge_canvas_with_images(self, images, channels: int = 3):
        columns = self._create_columns(images)
        columns = self._merge_img_in_columns(columns, channels)

        canvas = np.ones([self._column_height, self._img_width, channels], dtype=np.uint8) * 255
        for i, image in enumerate(columns):
            if image.shape[0] > canvas.shape[0] - self._g_spacing:
                image = image[: canvas.shape[0] - self._g_spacing, :]

            column_start = i * (self._column_width + self._g_spacing) + self._g_spacing
            column_end = column_start + self._column_width
            row_start = self._g_spacing
            row_end = canvas.shape[0]

            canvas[row_start:row_end, column_start:column_end] = image

        return canvas

    def _create_columns(self, images):
        num_images = len(images)
        image_heights = [image.shape[0] for image in images]

        one_big_column_height = (
            sum(image_heights[: -self._cols]) + (num_images - 2) * self._g_spacing
        )
        self._column_height = one_big_column_height // self._cols

        columns = []
        column_images = []
        current_height = 0

        for image, height in zip(images, image_heights):
            column_images.append(image)
            current_height += height + self._g_spacing

            if current_height > self._column_height:
                columns.append(column_images)

                column_images = []
                current_height = 0

        if len(columns) == self._cols:
            return columns
        return columns

    def _merge_img_in_columns(self, columns, channels: int = 3):
        combined_columns = []
        separator = np.ones((15, self._column_width, channels), dtype=np.uint8) * 255
        for column in columns:
            combined_images = []

            for image in column:
                combined_images.append(image)
                combined_images.append(separator)
            combined_images.pop()
            combined_image = np.vstack(combined_images)
            combined_columns.append(combined_image)

        return combined_columns

    def _resize_image(self, image, width):
        img_h, img_w = image.shape[:2]
        img_aspect_ratio = width / img_w
        img_w = int(width)
        img_h = int(img_aspect_ratio * img_h)

        image = sly.image.resize(image, (img_h, img_w))

        return image

    def _add_footer_with_logo(self, image):
        image2 = cv2.imread(self._footer_path, cv2.IMREAD_UNCHANGED)
        image2 = cv2.cvtColor(image2, cv2.COLOR_BGRA2RGBA)

        height1, width1 = image.shape[:2]
        height2, width2 = image2.shape[:2]

        if width1 != width2:
            scale_factor = width1 / width2
            height2, width2 = int(scale_factor * height2), width1
            image2 = sly.image.resize(image2, (height2, width2))

        alpha_channel = image2[:, :, 3] / 255.0

        x = 0
        y = height1 - height2

        region = image[y:height1, x : x + width2]
        image[y:height1, x : x + width2, :3] = (1 - alpha_channel[:, :, np.newaxis]) * region[
            :, :, :3
        ] + alpha_channel[:, :, np.newaxis] * image2[:, :, :3]

    def _overlay_images(self, bg, overlay, opacity):
        alpha = overlay[..., 3] * opacity / 255.0
        one_minus_alpha = 1 - alpha

        result_image = np.copy(bg)
        for c in range(3):
            result_image[..., c] = bg[..., c] * one_minus_alpha + overlay[..., c] * alpha
        return result_image

    def _save_video(self, videopath: str, frames):
        fourcc = cv2.VideoWriter_fourcc(*"VP90")
        height, width = frames[0].shape[:2]
        video_writer = cv2.VideoWriter(videopath, fourcc, 15, (width, height))

        with tqdm(desc="Saving video...", total=len(frames)) as vid_pbar:
            for frame in frames:
                if frame.shape[:2] != (height, width):
                    raise Exception("Not all frame sizes are not equal to each other.")
                frame = cv2.cvtColor(frame, cv2.COLOR_RGBA2BGR)
                video_writer.write(frame)
                vid_pbar.update(1)
        video_writer.release()
